import csv
from collections import namedtuple
from datetime import datetime as Datetime
from decimal import Decimal
from io import StringIO

from chellow.models import Session, Supply
from chellow.utils import HH, parse_mpan_core, to_ct, to_utc

from dateutil.relativedelta import relativedelta

from werkzeug.exceptions import BadRequest


READ_TYPE_MAP = {
    'A': 'A',
    'C': 'C',
    'D': 'D',
    'E': 'E',
    'F': 'F',
    'X': 'H',
    'I': 'I',
    'R': 'N',
    'O': 'O',
    'Q': 'Q',
    'S': 'S',
    'Z': 'Z',
}

SSC_MAP = {
    # 0326
    '20 0000 5855 970': {
        'Night': '00187',
        'Other': '00210',
        'Weekday': '00184'
    },
    '20 0000 6048 528': {
        'Night': '00187',
        'Other': '00210',
        'Weekday': '00184'
    },
    '20 0002 1909 377': {
        'Night': '00187',
        'Other': '00210',
        'Weekday': '00184'
    },
    '20 0002 2254 214': {
        'Night': '00187',
        'Other': '00210',
        'Weekday': '00184'
    },
    '20 0002 2326 515': {
        'Night': '00187',
        'Other': '00210',
        'Weekday': '00184'
    },
    '20 0002 2371 155': {
        'Night': '00187',
        'Other': '00210',
        'Weekday': '00184'
    },
    '20 0002 5282 171': {
        'Night': '00187',
        'Other': '00210',
        'Weekday': '00184'
    },
    '20 0002 5476 287': {
        'Night': '00187',
        'Other': '00210',
        'Weekday': '00184'
    },
    '20 0002 6228 157': {
        'Night': '00187',
        'Other': '00210',
        'Weekday': '00184'
    },
    '20 0002 6440 184': {
        'Night': '00187',
        'Other': '00210',
        'Weekday': '00184'
    },
    '20 0002 6419 768': {
        'Night': '00187',
        'Other': '00210',
        'Weekday': '00184'
    },
    # 0246
    '22 0001 3834 361': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    '22 0001 4442 321': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    '22 0002 1401 756': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    '22 0002 1427 578': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    '22 0002 1442 103': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    '22 0002 1502 222': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    '22 0002 1528 986': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    '22 0002 1540 176': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    '22 0002 1589 599': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    '22 0002 1823 142': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    '22 0002 1829 376': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    '22 0002 1839 541': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    '22 0002 1878 160': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    '22 0002 1394 727': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    # 0174
    '20 0002 6308 157': {
        'Day': '01072',
        'Night': '01071'
    },
    '20 0002 6467 511': {
        'Day': '01072',
        'Night': '01071'
    },
    '0008': {
        'Other': '00159',
    },
    '0065': {
        'Other': '00010',
    },
    '0036': {
        'Other': '00151',
    },
    '0037': {
        'Other': '00153',
    },
    '0151': {
        'Day': '00043',
        'Night': '00210',
    },
    '0154': {
        'Day': '00039',
        'Night': '00221',
    },
    '0174': {
        'Day': '01071',
        'Night': '01072'
    },
    '0179': {
        'Day': '01139',
        'Night': '01140'
    },
    '0184': {
        'Day': '01149',
        'Night': '01150'
    },
    '0186': {
        'Day': '01153',
        'Night': '01154'
    },
    '0242': {
        'Day': '00044',
        'Night': '00208'
    },
    '0244': {
        'Day': '00040',
        'Night': '00206'
    },
    '0246': {
        'Night': '00160',
        'Other': '00277',
        'Weekday': '00276',
    },
    '0252': {
        'Other': '00212',
    },
    '0265': {
        'Other': '00190',
    },
    '0319': {
        'Other': '00071',
        'Weekday': '00183',
    },
    '0320': {
        'Other': '00072',
        'Weekday': '00184',
    },
    '0322': {
        'Weekday': '01073',
        'Other': '01074',
    },
    '0326': {
        'Night': '00210',
        'Other': '00187',
        'Weekday': '00184',
    },
    '0393': {
        'Day': '00001'
    },
    '0428': {
        'Energy Charges': '00258',
        'Energy Charges 2': '00259',
    },
}


# None denotes a TPR-based charge

TITLE_MAP = {
    'Standing Charge': ('standing-gbp', 'standing-rate', 'standing-days'),
    'Climate Change Levy': ('ccl-gbp', 'ccl-rate', 'ccl-kwh'),
    'CCL': ('ccl-gbp', 'ccl-rate', 'ccl-kwh'),
    'Night': None,
    'Day': None,
    'Weekday': None,
    'Other': None,
    'Energy Charges': None,
}


BillElement = namedtuple(
    'BillElement', ['gbp', 'rate', 'cons', 'titles', 'desc'])


def _to_date(row, index):
    try:
        return to_utc(to_ct(Datetime.strptime(row[index], "%Y%m%d")))
    except ValueError as e:
        raise BadRequest(f"Can't parse the date at index {index}: {e}.")


class Parser():
    def __init__(self, f):
        self.reader = csv.reader(
            StringIO(str(f.read(), 'utf-8', errors='ignore')),
            skipinitialspace=True)
        self.line_number = None

    def make_raw_bills(self):
        raw_bills = []
        sess = Session()
        headers = {'sess': sess}

        blank_set = set(('',))
        for self.line_number, row in enumerate(self.reader):

            # skip blank lines
            if len(row) == 0 or set(row) == blank_set:
                continue

            try:
                bill = _process_line(row[0], row[1:], headers)
            except BadRequest as e:
                raise BadRequest(
                    f"Can't parse line number {self.line_number}: {row} : "
                    f"{e.description}")

            if bill is not None:
                raw_bills.append(bill)
        sess.close()
        return raw_bills


def _process_line(code, row, headers):
    if 'breakdown' in headers:
        headers['breakdown']['raw-lines'].append(row)

    if code == "INVOICE":
        _process_INVOICE(row, headers)

    elif code == "MPAN":
        _process_MPAN(row, headers)

    elif code == "VAT":
        return _process_VAT(row, headers)

    elif code == "READING":
        _process_READING(row, headers)

    elif code == "CHARGE":
        _process_CHARGE(row, headers)

    elif code == "SUMMARY":
        _process_SUMMARY(row, headers)

    else:
        raise BadRequest(f"The code {code} is not recognized.")


def _process_INVOICE(row, headers):
    sess = headers['sess']
    headers.clear()
    headers['kwh'] = Decimal('0')
    headers['reads'] = []
    headers['breakdown'] = {'raw-lines': []}
    headers['bill_elements'] = []
    headers['sess'] = sess

    headers['reference'] = row[0]
    bill_type_raw = row[3]
    headers['bill_type_code'] = 'W' if bill_type_raw == '' else bill_type_raw
    headers['issue_date'] = _to_date(row, 14)

    headers['start_date'] = _to_date(row, 17)
    headers['finish_date'] = _to_date(row, 18) + relativedelta(days=1) - HH
    headers['net'] = Decimal('0.00') + Decimal(row[19])
    headers['vat'] = Decimal('0.00') + Decimal(row[20])
    headers['gross'] = Decimal('0.00') + Decimal(row[21])
    headers['account'] = row[26]


def _process_MPAN(row, headers):
    dno = row[3]
    unique = row[4].zfill(10)
    check_digit = row[5]

    headers['mpan_core'] = parse_mpan_core(''.join([dno, unique, check_digit]))


def _process_VAT(rows, headers):
    sess = headers['sess']
    mpan_core = headers['mpan_core']
    start_date = headers['start_date']
    reads = headers['reads']
    supply = Supply.get_by_mpan_core(sess, mpan_core)
    era = supply.find_era_at(sess, start_date)
    bill_elements = []
    if era is None:
        era = supply.find_last_era(sess)
    if era is not None and era.ssc is not None:
        try:
            ssc_lookup = era.imp_mpan_core
            tpr_map = SSC_MAP[ssc_lookup]
        except KeyError:
            ssc_lookup = era.ssc.code
            try:
                tpr_map = SSC_MAP[ssc_lookup]
            except KeyError:
                raise BadRequest(
                    "The SSC " + ssc_lookup + " isn't in the SSC_MAP.")

        for read in reads:
            desc = read['tpr_code']
            try:
                read['tpr_code'] = tpr_map[desc]
            except KeyError:
                raise BadRequest(
                    "The description " + desc + " isn't in the SSC_MAP "
                    "for the SSC " + ssc_lookup + ".")

        for el in headers['bill_elements']:
            if el.titles is None:
                try:
                    tpr = tpr_map[el.desc]
                except KeyError:
                    raise BadRequest(
                        f"The billing element description {el.desc} isn't in "
                        f"the SSC_MAP for the SSC {ssc_lookup}.")

                titles = (tpr + '-gbp', tpr + '-rate', tpr + '-kwh')
            else:
                titles = el.titles

            bill_elements.append(
                BillElement(
                    gbp=el.gbp, titles=titles, rate=el.rate, cons=el.cons,
                    desc=None))
    else:
        for read in reads:
            read['tpr_code'] = '00001'

        for el in headers['bill_elements']:
            if el.titles is None:
                des = el.desc
                titles = (des + '-kwh', des + '-rate', des + '-gbp')
            else:
                titles = el.titles

            bill_elements.append(
                BillElement(
                    gbp=el.gbp, titles=titles, rate=el.rate, cons=el.cons,
                    desc=None))

    breakdown = headers['breakdown']
    for bill_el in bill_elements:
        eln_gbp, eln_rate, eln_cons = bill_el.titles
        breakdown[eln_gbp] = bill_el.gbp
        rate = bill_el.rate
        if eln_rate is not None and rate is not None:
            try:
                rates = breakdown[eln_rate]
            except KeyError:
                rates = breakdown[eln_rate] = set()

            rates.add(rate)

        cons = bill_el.cons
        if eln_cons is not None and cons is not None:
            breakdown[eln_cons] = cons

    return {
        'kwh': headers['kwh'],
        'reference': headers['reference'],
        'mpan_core': mpan_core,
        'issue_date': headers['issue_date'],
        'account': headers['account'],
        'start_date': start_date,
        'finish_date': headers['finish_date'],
        'net': headers['net'],
        'vat': headers['vat'],
        'gross': headers['gross'],
        'breakdown': breakdown,
        'reads': reads,
        'bill_type_code': headers['bill_type_code'],
    }


def _process_READING(row, headers):
    coefficient = Decimal(row[2])
    msn = row[4]
    tpr_code_raw = row[6]

    prev_read_date = _to_date(row, 11) + relativedelta(days=1) - HH
    prev_reading_value = Decimal(row[12])
    try:
        prev_read_type = READ_TYPE_MAP[row[13]]
    except KeyError as e:
        raise BadRequest(
            f"The previous register read type isn't recognized {e}.")

    pres_read_date = _to_date(row, 14) + relativedelta(days=1) - HH
    pres_reading_value = Decimal(row[15])
    try:
        pres_read_type = READ_TYPE_MAP[row[16]]
    except KeyError as e:
        raise BadRequest(
            f"The present register read type isn't recognized {e}.")

    units_raw = row[18]
    if units_raw == 'kW':
        units = 'kW'
        tpr_code = None
    elif units_raw == 'kVA':
        units = 'kVA'
        tpr_code = None
    elif units_raw == 'KWH':
        units = 'kWh'
        tpr_code = tpr_code_raw
    else:
        raise BadRequest(f"Units {units_raw} not recognized.")

    headers['reads'].append(
        {
            'msn': msn,
            'mpan': '',
            'coefficient': coefficient,
            'units': units,
            'tpr_code': tpr_code,
            'prev_date': prev_read_date,
            'prev_value': prev_reading_value,
            'prev_type_code': prev_read_type,
            'pres_date': pres_read_date,
            'pres_value': pres_reading_value,
            'pres_type_code': pres_read_type
        }
    )


def _process_CHARGE(row, headers):
    desc = row[6]

    try:
        titles = TITLE_MAP[desc]
    except KeyError:
        raise BadRequest(
            f"Can't find the description {desc} in the TITLE_MAP.")

    rate = Decimal(row[20])

    if desc == 'Standing Charge':
        start_date = _to_date(row, 23)
        finish_date = _to_date(row, 24)
        consumption = Decimal((finish_date - start_date).days + 1)
    else:
        consumption = Decimal(row[22])

    if titles is None:
        headers['kwh'] += consumption

    gbp = Decimal('0.00') + Decimal(row[25])

    if desc == 'Energy Charges':
        headers['bill_elements'].append(
            BillElement(
                gbp=(gbp / 2), rate=rate, cons=(consumption / 2),
                titles=titles, desc=desc))
        headers['bill_elements'].append(
            BillElement(
                gbp=(gbp / 2), rate=rate, cons=(consumption / 2),
                titles=titles, desc='Energy Charges 2'))
    else:
        headers['bill_elements'].append(
            BillElement(
                gbp=gbp, rate=rate, cons=consumption, titles=titles,
                desc=desc))


def _process_SUMMARY(row, headers):
    pass



import collections
import threading
import datetime
import traceback
import urllib.parse
import http.client
from chellow.models import (
    Contract, RateScript, get_non_core_contract_id, Session)
from chellow.utils import HH, hh_format, utc_datetime_now, to_utc, to_ct
import xlrd
from werkzeug.exceptions import BadRequest
import atexit
from zish import loads
from datetime import timedelta as Timedelta


ELEXON_PORTAL_SCRIPTING_KEY_KEY = 'elexonportal_scripting_key'


def key_format(dt):
    return dt.strftime("%d %H:%M")


def hh(data_source):
    for h in data_source.hh_data:
        try:
            sbp, ssp = data_source.caches['system_price'][h['start-date']]
        except KeyError:
            try:
                system_price_cache = data_source.caches['system_price']
            except KeyError:
                system_price_cache = data_source.caches['system_price'] = {}

            db_id = get_non_core_contract_id('system_price')
            h_start = h['start-date']
            rates = data_source.hh_rate(db_id, h_start)['gbp_per_nbp_mwh']

            try:
                try:
                    rdict = rates[key_format(h_start)]
                except KeyError:
                    rdict = rates[key_format(h_start - Timedelta(days=3))]
                sbp = float(rdict['sbp'] / 1000)
                ssp = float(rdict['ssp'] / 1000)
                system_price_cache[h_start] = (sbp, ssp)
            except KeyError:
                raise BadRequest(
                    "For the System Price rate script at " +
                    hh_format(h_start) + " the rate cannot be found.")
            except TypeError:
                raise BadRequest(
                    "For the System Price rate script at " +
                    hh_format(h_start) +
                    " the rate 'rates_gbp_per_mwh' has the problem: " +
                    traceback.format_exc())

        h['sbp'] = sbp
        h['sbp-gbp'] = h['nbp-kwh'] * sbp

        h['ssp'] = ssp
        h['ssp-gbp'] = h['nbp-kwh'] * ssp


system_price_importer = None


class SystemPriceImporter(threading.Thread):
    def __init__(self):
        super(SystemPriceImporter, self).__init__(name="System Price Importer")
        self.lock = threading.RLock()
        self.messages = collections.deque()
        self.stopped = threading.Event()
        self.going = threading.Event()
        self.PROXY_HOST_KEY = 'proxy.host'
        self.PROXY_PORT_KEY = 'proxy.port'

    def stop(self):
        self.stopped.set()
        self.going.set()
        self.join()

    def go(self):
        self.going.set()

    def is_locked(self):
        if self.lock.acquire(False):
            self.lock.release()
            return False
        else:
            return True

    def log(self, message):
        self.messages.appendleft(
            utc_datetime_now().strftime("%Y-%m-%d %H:%M:%S") + " - " + message)
        if len(self.messages) > 100:
            self.messages.pop()

    def run(self):
        while not self.stopped.isSet():
            if self.lock.acquire(False):
                sess = book = sbp_sheet = ssp_sheet = None
                try:
                    sess = Session()
                    self.log("Starting to check System Prices.")
                    # ct_tz = pytz.timezone('Europe/London')
                    contract = Contract.get_non_core_by_name(
                        sess, 'system_price')
                    contract_props = contract.make_properties()

                    if contract_props.get('enabled', False):
                        for rscript in sess.query(RateScript).filter(
                                RateScript.contract == contract).order_by(
                                RateScript.start_date.desc()):
                            ns = loads(rscript.script)
                            rates = ns['gbp_per_nbp_mwh']
                            if len(rates) == 0:
                                fill_start = rscript.start_date
                                break
                            elif rates[
                                    key_format(
                                        rscript.finish_date)]['run'] == 'DF':
                                fill_start = rscript.finish_date + HH
                                break

                        config = Contract.get_non_core_by_name(
                            sess, 'configuration')
                        config_props = config.make_properties()

                        scripting_key = config_props.get(
                            ELEXON_PORTAL_SCRIPTING_KEY_KEY)
                        if scripting_key is None:
                            raise BadRequest(
                                "The property " +
                                ELEXON_PORTAL_SCRIPTING_KEY_KEY +
                                " cannot be found in the configuration "
                                "properties.")
                        url_str = contract_props['url'] + \
                            'file/download/BESTVIEWPRICES_FILE?key=' + \
                            scripting_key

                        self.log(
                            "Downloading from " + url_str +
                            " and extracting data from " +
                            hh_format(fill_start))

                        url = urllib.parse.urlparse(url_str)
                        if url.scheme == 'https':
                            conn = http.client.HTTPSConnection(
                                url.hostname, url.port)
                        else:
                            conn = http.client.HTTPConnection(
                                url.hostname, url.port)
                        conn.request("GET", url.path + '?' + url.query)

                        res = conn.getresponse()
                        self.log(
                            "Received " + str(res.status) + " " + res.reason)
                        data = res.read()
                        book = xlrd.open_workbook(file_contents=data)
                        sbp_sheet = book.sheet_by_index(1)
                        ssp_sheet = book.sheet_by_index(2)

                        sp_months = []
                        sp_month = None
                        for row_index in range(1, sbp_sheet.nrows):
                            sbp_row = sbp_sheet.row(row_index)
                            ssp_row = ssp_sheet.row(row_index)
                            raw_date = datetime.datetime(
                                *xlrd.xldate_as_tuple(
                                    sbp_row[0].value, book.datemode))
                            hh_date_ct = to_ct(raw_date)
                            hh_date = to_utc(hh_date_ct)
                            run_code = sbp_row[1].value
                            for col_idx in range(2, 52):
                                if hh_date >= fill_start:
                                    sbp_val = sbp_row[col_idx].value
                                    if sbp_val != '':
                                        if hh_date.day == 1 and \
                                                hh_date.hour == 0 and \
                                                hh_date.minute == 0:
                                            sp_month = {}
                                            sp_months.append(sp_month)
                                        ssp_val = ssp_row[col_idx].value
                                        sp_month[hh_date] = {
                                            'run': run_code,
                                            'sbp': sbp_val, 'ssp': ssp_val}
                                hh_date += HH
                        self.log("Successfully extracted data.")
                        last_date = sorted(sp_months[-1].keys())[-1]
                        if last_date.month == (last_date + HH).month:
                            del sp_months[-1]
                        if 'limit' in contract_props:
                            sp_months = sp_months[0:1]
                        for sp_month in sp_months:
                            sorted_keys = sorted(sp_month.keys())
                            month_start = sorted_keys[0]
                            month_finish = sorted_keys[-1]
                            rs = sess.query(RateScript).filter(
                                RateScript.contract == contract,
                                RateScript.start_date == month_start).first()
                            if rs is None:
                                self.log(
                                    "Adding a new rate script starting at " +
                                    hh_format(month_start) + ".")

                                latest_rs = sess.query(RateScript).filter(
                                    RateScript.contract == contract).\
                                    order_by(RateScript.start_date.desc()). \
                                    first()

                                contract.update_rate_script(
                                    sess, latest_rs, latest_rs.start_date,
                                    month_finish, loads(latest_rs.script))
                                rs = contract.insert_rate_script(
                                    sess, month_start, {})
                                sess.flush()
                            script = {
                                'gbp_per_nbp_mwh': dict(
                                    (key_format(k), v)
                                    for k, v in sp_month.items())}
                            self.log(
                                "Updating rate script starting at " +
                                hh_format(month_start) + ".")
                            contract.update_rate_script(
                                sess, rs, rs.start_date, rs.finish_date,
                                script)
                            sess.commit()
                    else:
                        self.log(
                            "The automatic importer is disabled. To "
                            "enable it, edit the contract properties to "
                            "set 'enabled' to True.")

                except BaseException:
                    self.log("Outer problem " + traceback.format_exc())
                    sess.rollback()
                finally:
                    book = sbp_sheet = ssp_sheet = None
                    self.lock.release()
                    self.log("Finished checking System Price rates.")
                    if sess is not None:
                        sess.close()

            self.going.wait(24 * 60 * 60)
            self.going.clear()


def get_importer():
    return system_price_importer


def startup():
    global system_price_importer
    system_price_importer = SystemPriceImporter()
    system_price_importer.start()


@atexit.register
def shutdown():
    if system_price_importer is not None:
        system_price_importer.stop()



from decimal import Decimal, InvalidOperation
from dateutil.relativedelta import relativedelta
from collections import defaultdict
from chellow.edi_lib import EdiParser, SEGMENTS
from chellow.utils import HH, to_utc, to_ct
from io import StringIO
from werkzeug.exceptions import BadRequest
from datetime import datetime as Datetime


read_type_map = {
    '00': 'N', '09': 'N3', '04': 'C', '02': 'E', '11': 'E3', '01': 'EM',
    '03': 'W', '06': 'X', '05': 'CP', '12': 'IF'}


TCOD_MAP = {
    '030025': ('ccl-gbp', 'ccl-rate', 'ccl-kwh'),
    '345065': (
        'summer-weekend-gbp', 'summer-weekend-rate', 'summer-weekend-kwh'),
    '350293': ('capacity-gbp', 'capacity-rate', 'capacity-kwh'),
    '425779': ('ro-gbp', 'ro-rate', 'ro-kwh'),
    '534342': ('reconciliation-gbp', None, None),
    '584867': ('aahedc-gbp', 'aahedc-rate', 'aahedc-kwh'),
    '989534': ('bsuos-gbp', 'bsuos-rate', 'bsuos-kwh'),
    '579387': ('capacity-gbp', 'capacity-rate', 'capacity-kwh'),
    '558147': ('capacity-gbp', 'capacity-rate', 'capacity-kwh'),
    '066540': ('ccl-gbp', 'ccl-rate', 'ccl-kwh'),
    '378809': ('cfd-fit-gbp', 'cfd-fit-rate', 'cfd-fit-kwh'),
    '281170': ('cfd-fit-gbp', 'cfd-fit-rate', 'cfd-fit-kwh'),
    '574015': ('cfd-fit-gbp', 'cfd-fit-rate', 'cfd-fit-kwh'),
    '068476':  ('day-gbp', 'day-rate', 'day-kwh'),
    '219182': (
        'duos-availability-gbp', 'duos-availability-rate',
        'duos-availability-kva'),
    '144424': (
        'duos-excess-availability-gbp', 'duos-excess-availability-rate',
        'duos-excess-availability-kva'),
    '986159': ('duos-fixed-gbp', 'duos-fixed-rate', 'duos-fixed-days'),
    '838286': (
        'duos-reactive-gbp', 'duos-reactive-rate', 'duos-reactive-kvarh'),
    '242643': ('duos-fixed-gbp', 'duos-fixed-rate', 'duos-fixed-days'),
    '257303': ('duos-red-gbp', 'duos-red-rate', 'duos-red-kwh'),
    '257304': ('duos-amber-gbp', 'duos-amber-rate', 'duos-amber-kwh'),
    '257305': ('duos-green-gbp', 'duos-green-rate', 'duos-green-kwh'),
    '661439': ('duos-red-gbp', 'duos-red-rate', 'duos-red-kwh'),
    '661440': ('duos-amber-gbp', 'duos-amber-rate', 'duos-red-kwh'),
    '309707': ('fit-gbp', 'fit-rate', 'fit-kwh'),
    '994483': ('reconciliation-gbp', None, None),
    '544936': ('meter-rental-gbp', 'meter-rental-rate', 'meter-rental-days'),
    '265091': ('night-gbp', 'night-rate', 'night-kwh'),
    '483457': ('peak-gbp', 'peak-rate', 'peak-kwh'),
    '975901': ('peak-shoulder-gbp', 'peak-shoulder-rate', 'peak-shoulder-kwh'),
    '378246': ('ro-gbp', 'ro-rate', 'ro-kwh'),
    '632209': ('summer-night-gbp', 'summer-night-rate', 'summer-night-kwh'),
    '663682': (
        'summer-weekday-gbp', 'summer-weekday-rate', 'summer-weekday-kwh'),
    '299992': (
        'summer-weekend-gbp', 'summer-weekend-rate', 'summer-weekend-kwh'),
    '447769': ('triad-gbp', 'triad-rate', 'triad-kw'),
    '647721': ('triad-gbp', 'triad-rate', 'triad-kw'),
    '276631': ('triad-gbp', 'triad-rate', 'triad-kw'),
    '220894': ('winter-night-gbp', 'winter-night-rate', 'winter-night-kwh'),
    '264929': (
        'winter-weekday-gbp', 'winter-weekday-rate', 'winter-weekday-kwh'),
    '638187': (
        'winter-weekend-gbp', 'winter-weekend-rate', 'winter-weekend-kwh'),
    '700285': ('duo-fixed-gbp', 'duos-fixed-rate', 'duos-fixed-days'),
}


def _to_date(component):
    return to_utc(to_ct(Datetime.strptime(component, "%y%m%d")))


def _to_decimal(components, divisor=None):
    comp_0 = components[0]

    try:
        result = Decimal(comp_0)
    except InvalidOperation as e:
        raise BadRequest(
            "Can't parse '" + str(comp_0) + "' of " + str(components) +
            " as a decimal:" + str(e))

    if len(components) > 1 and components[-1] == "R":
        result *= Decimal("-1")

    if divisor is not None:
        result /= Decimal(divisor)

    return result


def _find_elements(code, elements):
    segment_name = code + elements[1][0] if code == 'CCD' else code
    elem_codes = [m['code'] for m in SEGMENTS[segment_name]['elements']]
    return dict(zip(elem_codes, elements))


class Parser():
    def __init__(self, f):
        self.parser = EdiParser(
            StringIO(str(f.read(), 'utf-8', errors='ignore')))
        self.line_number = None

    def make_raw_bills(self):
        raw_bills = []
        headers = {}
        for self.line_number, code in enumerate(self.parser):
            elements = _find_elements(code, self.parser.elements)
            line = self.parser.line
            try:
                bill = _process_segment(code, elements, line, headers)
            except BadRequest as e:
                raise BadRequest(
                    "Can't parse the line: " + line + " :" + e.description)
            if bill is not None:
                raw_bills.append(bill)
        return raw_bills


def _process_segment(code, elements, line, headers):
    if code == "BCD":
        issue_date = _to_date(elements['IVDT'][0])
        reference = elements['INVN'][0]
        bill_type_code = elements['BTCD'][0]

        sumo = elements['SUMO']
        start_date = _to_date(sumo[0])
        finish_date = _to_date(sumo[1]) + relativedelta(days=1) - HH

        headers['issue_date'] = issue_date
        headers['bill_type_code'] = bill_type_code
        headers['reference'] = reference

    elif code == "MHD":
        message_type = elements['TYPE'][0]
        if message_type == "UTLBIL":
            pass

    elif code == "CCD":
        consumption_charge_indicator = elements['CCDE'][0]

        if consumption_charge_indicator == "1":
            pres_read_date = _to_date(elements['PRDT'][0]) + relativedelta(
                days=1) - HH

            prev_read_date = _to_date(elements['PVDT'][0]) + relativedelta(
                days=1) - HH

            m = elements['MLOC'][0]
            mpan = ' '.join(
                (m[13:15], m[15:18], m[18:], m[:2], m[2:6], m[6:10], m[10:13]))

            prrd = elements['PRRD']
            pres_read_type = read_type_map[prrd[1]]
            prev_read_type = read_type_map[prrd[3]]

            coefficient = Decimal(elements['ADJF'][1]) / Decimal(100000)
            pres_reading_value = Decimal(prrd[0])
            prev_reading_value = Decimal(prrd[2])
            msn = elements['MTNR'][0]
            tpr_code = elements['TMOD'][0]
            if tpr_code == 'kW':
                units = 'kW'
                tpr_code = None
            elif tpr_code == 'kVA':
                units = 'kVA'
                tpr_code = None
            else:
                units = 'kWh'
                kwh = _to_decimal(elements['CONS'], '1000')

            try:
                reads = headers['reads']
            except KeyError:
                reads = headers['reads'] = []

            reads.append(
                {
                    'msn': msn, 'mpan': mpan,
                    'coefficient': coefficient, 'units': units,
                    'tpr_code': tpr_code, 'prev_date': prev_read_date,
                    'prev_value': prev_reading_value,
                    'prev_type_code': prev_read_type,
                    'pres_date': pres_read_date,
                    'pres_value': pres_reading_value,
                    'pres_type_code': pres_read_type
                }
            )

        elif consumption_charge_indicator == "2":
            breakdown = defaultdict(int, {'raw-lines': line})

            element_code = elements['TCOD'][0]
            try:
                eln_gbp, eln_rate, eln_cons = TCOD_MAP[element_code]
            except KeyError:
                raise BadRequest(
                    "Can't find the element code " + element_code +
                    " in the TCOD_MAP.")

            m = elements['MLOC'][0]
            mpan_core = ' '.join((m[:2], m[2:6], m[6:10], m[10:]))

            cons = elements['CONS']
            kwh = Decimal('0')
            if eln_cons is not None and len(cons[0]) > 0:
                el_cons = _to_decimal(cons, '1000')
                breakdown[eln_cons] = kwh = el_cons

            if eln_rate is not None:
                rate = _to_decimal(elements['BPRI'], '100000')
                breakdown[eln_rate] = [rate]

            start_date = _to_date(elements['CSDT'][0])
            headers['bill_start_date'] = start_date

            finish_date = _to_date(elements['CEDT'][0]) - HH
            headers['bill_finish_date'] = finish_date

            if 'CTOT' in elements:
                net = Decimal('0.00') + _to_decimal(elements['CTOT'], '100')
            else:
                net = Decimal('0.00')

            breakdown[eln_gbp] = net

            headers['mpan_core'] = mpan_core

            try:
                reads = headers['reads']
                del headers['reads'][:]
            except KeyError:
                reads = []

            return {
                'bill_type_code': headers['bill_type_code'],
                'reference': headers['reference'] + '_' + eln_gbp[:-4],
                'issue_date': headers['issue_date'],
                'mpan_core': mpan_core,
                'account': mpan_core,
                'start_date': start_date,
                'finish_date': finish_date,
                'kwh': kwh if eln_gbp == 'ro-gbp' else Decimal('0'),
                'net': net,
                'vat': Decimal('0.00'),
                'gross': net,
                'breakdown': breakdown,
                'reads': reads
            }

        elif consumption_charge_indicator == '3':
            breakdown = defaultdict(int, {'raw-lines': line})

            element_code = elements['TCOD'][0]
            try:
                eln_gbp, eln_rate, eln_cons = TCOD_MAP[element_code]
            except KeyError:
                raise BadRequest(
                    "Can't find the element code " + element_code +
                    " in the TCOD_MAP.")

            m = elements['MLOC'][0]
            mpan_core = ' '.join((m[:2], m[2:6], m[6:10], m[10:]))

            cons = elements['CONS']
            if eln_cons is not None and len(cons[0]) > 0:
                el_cons = _to_decimal(cons, '1000')
                breakdown[eln_cons] = kwh = el_cons
            else:
                kwh = Decimal('0')

            if eln_rate is not None:
                rate = _to_decimal(elements['BPRI'], '100000')
                breakdown[eln_rate] = [rate]

            start_date = _to_date(elements['CSDT'][0])
            headers['bill_start_date'] = start_date

            finish_date = _to_date(elements['CEDT'][0]) - HH
            headers['bill_finish_date'] = finish_date

            if 'CTOT' in elements:
                net = Decimal('0.00') + _to_decimal(elements['CTOT'], '100')
            else:
                net = Decimal('0.00')

            breakdown[eln_gbp] = net

            headers['mpan_core'] = mpan_core

            try:
                reads = headers['reads']
                del headers['reads'][:]
            except KeyError:
                reads = []

            return {
                'bill_type_code': headers['bill_type_code'],
                'issue_date': headers['issue_date'],
                'reference': headers['reference'],
                'mpan_core': mpan_core,
                'account': mpan_core,
                'start_date': start_date,
                'finish_date': finish_date,
                'net': net,
                'kwh': kwh if eln_gbp == 'ro-gbp' else Decimal('0'),
                'vat': Decimal('0.00'),
                'gross': net,
                'breakdown': breakdown,
                'reads': reads
            }

        elif consumption_charge_indicator == '4':
            breakdown = defaultdict(int, {'raw-lines': line})

            element_code = elements['TCOD'][0]
            try:
                eln_gbp, eln_rate, eln_cons = TCOD_MAP[element_code]
            except KeyError:
                raise BadRequest(
                    "Can't find the element code " + element_code +
                    " in the TCOD_MAP.")

            m = elements['MLOC'][0]
            mpan_core = ' '.join((m[:2], m[2:6], m[6:10], m[10:]))

            cons = elements['CONS']
            if eln_cons is not None and len(cons[0]) > 0:
                el_cons = _to_decimal(cons, '1000')
                breakdown[eln_cons] = kwh = el_cons

            if eln_rate is not None:
                rate = _to_decimal(elements['BPRI'], '100000')
                breakdown[eln_rate] = [rate]

            start_date = _to_date(elements['CSDT'][0])
            headers['bill_start_date'] = start_date

            finish_date = _to_date(elements['CEDT'][0]) - HH
            headers['bill_finish_date'] = finish_date

            if 'CTOT' in elements:
                net = Decimal('0.00') + _to_decimal(elements['CTOT'], '100')
            else:
                net = Decimal('0.00')

            breakdown[eln_gbp] = net

            headers['mpan_core'] = mpan_core

            try:
                reads = headers['reads']
                del headers['reads'][:]
            except KeyError:
                reads = []

            return {
                'kwh': kwh if eln_gbp == 'ro-gbp' else Decimal('0.00'),
                'reference': headers['reference'] + '_' + eln_gbp[:-4],
                'issue_date': headers['issue_date'],
                'mpan_core': mpan_core,
                'account': mpan_core,
                'start_date': start_date,
                'finish_date': finish_date,
                'net': net,
                'vat': Decimal('0.00'),
                'gross': net,
                'breakdown': breakdown,
                'reads': reads,
                'bill_type_code': headers['bill_type_code'],
            }

    elif code == "MTR":
        headers.clear()

    elif code == "MAN":
        pass

    elif code == "VAT":
        vat = Decimal('0.00') + _to_decimal(elements['UVTT'], '100')

        return {
            'bill_type_code': headers['bill_type_code'],
            'account': headers['mpan_core'],
            'mpan_core': headers['mpan_core'],
            'reference': headers['reference'] + '_vat',
            'issue_date': headers['issue_date'],
            'start_date': headers['bill_start_date'],
            'finish_date': headers['bill_finish_date'],
            'kwh': Decimal('0.00'),
            'net': Decimal('0.00'),
            'vat': vat,
            'gross': vat,
            'breakdown': {},
            'reads': []
        }


def _get_element(elements, index, line):
    try:
        return elements[index]
    except IndexError:
        raise BadRequest(
            "The index " + str(index) + " is past the end of the elements " +
            str(elements) + " and line " + line + ".")



import csv
import decimal
from decimal import Decimal

from chellow.utils import HH, utc_datetime, utc_datetime_now

from werkzeug.exceptions import BadRequest

from xlrd import open_workbook


def get_value(row, title_row, name):
    try:
        idx = title_row.index(name)
    except ValueError:
        return None

    try:
        val = row[idx].value
    except IndexError:
        raise BadRequest(
            f"For the name '{name}', the index is {idx} which is beyond the "
            f"end of the row.")

    if isinstance(val, str):
        return val.strip()
    else:
        return val


def get_dec(row, title_row, name):
    val = get_value(row, title_row, name)

    try:
        return val if val is None else Decimal(str(val))
    except decimal.InvalidOperation:
        return None


def get_rate(row, title_row, name):
    raw = get_dec(row, title_row, name)
    return None if raw is None else [raw / Decimal('100')]


def get_date_ct(row, title_row, name):
    s = get_value(row, title_row, name)
    nums = list(map(int, s.split('-')))
    return utc_datetime(nums[0], nums[1], nums[2])


def _parse_row(row, row_index, datemode, title_row):
    titles = [c.value for c in title_row]
    bill_start_date = get_date_ct(row, titles, 'BILL START DATE')
    bill_finish_date = get_date_ct(row, titles, 'BILL END DATE') + (HH * 47)
    issue_date = utc_datetime_now()

    aahedc_kwh = get_dec(row, titles, 'LVY-AAHEDC-ALL USE KWH')

    vals = [
        (
            'cap-mech-ob-kwh',
            get_dec(row, titles, 'LVY-CMLOB-ALL USE KWH')),
        ('cap-mech-ob-rate', get_dec(row, titles, 'LVY-CMLOB-ALL RATE P/KWH')),
        ('cap-mech-ob-gbp', get_dec(row, titles, 'LVY-CMLOB-ALL COST GBP')),
        ('cap-mech-op-kwh', get_dec(row, titles, 'LVY-CMLOP-ALL USE KWH')),
        ('cap-mech-op-rate', get_dec(row, titles, 'LVY-CMLOP-ALL RATE P/KWH')),
        ('cap-mech-op-gbp', get_dec(row, titles, 'LVY-CMLOP-ALL COST GBP')),
        ('duos-fixed-days', get_dec(row, titles, 'DUOS-STND-SITE USE DAY')),
        (
            'duos-availability-kva-days',
            [get_dec(row, titles, 'DUOS-AVAIL-AV USE KVADAY')]
        ),
        (
            'duos-availability-rate',
            get_rate(row, titles, 'DUOS-AVAIL-AV RATE P/KVADAY')
        ),
        (
            'duos-availability-gbp',
            get_dec(row, titles, 'DUOS-AVAIL-AV COST GBP')),
        (
            'duos-fixed-rate',
            get_rate(row, titles, 'DUOS-STND-SITE RATE P/DAY')
        ),
        ('duos-fixed-gbp', get_dec(row, titles, 'DUOS-STND-SITE COST GBP')),
        ('duos-amber-kwh', get_dec(row, titles, 'DUOS-UNIT-AMBER USE KWH')),
        (
            'duos-amber-rate',
            get_rate(row, titles, 'DUOS-UNIT-AMBER RATE P/KWH')
        ),
        ('duos-amber-gbp', get_dec(row, titles, 'DUOS-UNIT-AMBER COST GBP')),
        ('duos-green-kwh', get_dec(row, titles, 'DUOS-UNIT-GREEN USE KWH')),
        (
            'duos-green-rate',
            get_rate(row, titles, 'DUOS-UNIT-GREEN RATE P/KWH')
        ),
        ('duos-green-gbp', get_dec(row, titles, 'DUOS-UNIT-GREEN COST GBP')),
        ('duos-red-kwh', get_dec(row, titles, 'DUOS-UNIT-RED USE KWH')),
        ('duos-red-rate', get_rate(row, titles, 'DUOS-UNIT-RED RATE P/KWH')),
        ('duos-red-gbp', get_dec(row, titles, 'DUOS-UNIT-RED COST GBP')),
        ('aahedc-kwh', aahedc_kwh),
        ('aahedc-rate', get_rate(row, titles, 'LVY-AAHEDC-ALL RATE P/KWH')),
        ('aahedc-gbp', get_dec(row, titles, 'LVY-AAHEDC-ALL COST GBP')),
        ('bsuos-nbp-kwh', get_dec(row, titles, 'LVY-BSUOS-ALL USE KWH')),
        ('bsuos-rate', get_rate(row, titles, 'LVY-BSUOS-ALL RATE P/KWH')),
        ('bsuos-gbp', get_dec(row, titles, 'LVY-BSUOS-ALL COST GBP')),
        ('ccl-kwh', get_dec(row, titles, 'LVY-CCL-ALL USE KWH')),
        ('ccl-rate', get_rate(row, titles, 'LVY-CCL-ALL RATE P/KWH')),
        ('ccl-gbp', get_dec(row, titles, 'LVY-CCL-ALL COST GBP')),
        ('cfdob-kwh', get_dec(row, titles, 'LVY-CFDOB-ALL USE KWH')),
        ('cfdob-rate', get_rate(row, titles, 'LVY-CFDOB-ALL RATE P/KWH')),
        ('cfdob-gbp', get_dec(row, titles, 'LVY-CFDOB-ALL COST GBP')),
        ('cfdop-kwh', get_dec(row, titles, 'LVY-CFDOP-ALL USE KWH')),
        ('cfdop-rate', get_rate(row, titles, 'LVY-CFDOP-ALL RATE P/KWH')),
        ('cfdop-gbp', get_dec(row, titles, 'LVY-CFDOP-ALL COST GBP')),
        ('fit-kwh', get_dec(row, titles, 'LVY-FIT-ALL USE KWH')),
        ('fit-rate', get_rate(row, titles, 'LVY-FIT-ALL RATE P/KWH')),
        ('fit-gbp', get_dec(row, titles, 'LVY-FIT-ALL COST GBP')),
        ('ro-kwh', get_dec(row, titles, 'LVY-RO-ALL USE KWH')),
        ('ro-rate', get_rate(row, titles, 'LVY-RO-ALL RATE P/KWH')),
        ('ro-gbp', get_dec(row, titles, 'LVY-RO-ALL COST GBP')),
        ('summer-kwh', get_dec(row, titles, 'NRG-UNIT-SUMMER USE KWH')),
        ('summer-rate', get_rate(row, titles, 'NRG-UNIT-SUMMER RATE P/KWH')),
        ('summer-gbp', get_dec(row, titles, 'NRG-UNIT-SUMMER COST GBP')),
        ('winter-kwh', get_dec(row, titles, 'NRG-UNIT-WINTER USE KWH')),
        ('winter-rate', get_rate(row, titles, 'NRG-UNIT-WINTER RATE P/KWH')),
        ('winter-gbp', get_dec(row, titles, 'NRG-UNIT-WINTER COST GBP')),
        ('admin-months', get_dec(row, titles, 'SYS-ADMIN-ALL USE MO')),
        ('admin-rate', [get_dec(row, titles, 'SYS-ADMIN-ALL RATE GBP/MO')]),
        ('admin-gbp', get_dec(row, titles, 'SYS-ADMIN-ALL COST GBP')),
        ('triad-days', get_dec(row, titles, 'TUOS-TRIAD-SITE USE DAY')),
        ('triad-rate', [get_dec(row, titles, 'TUOS-TRIAD-SITE RATE GBP/DAY')]),
        ('triad-gbp', get_dec(row, titles, 'TUOS-TRIAD-SITE COST GBP')),
    ]

    bd = {
        'raw_lines': [str(title_row), str(row)],
    }

    for k, v in vals:
        if v is not None and v != [None]:
            bd[k] = v

    kwh = Decimal('0.00') if aahedc_kwh is None else aahedc_kwh
    net_gbp = Decimal('0.00')
    net_gbp += sum(v for k, v in bd.items() if k.endswith('-gbp'))
    vat_gbp = Decimal('0.00')
    reference = issue_date.strftime('%Y%m%dT%H%M') + '_' + str(row_index + 1)

    return {
        'bill_type_code': 'N',
        'kwh': kwh,
        'vat': Decimal('0.00'),
        'net': net_gbp,
        'gross': net_gbp + vat_gbp,
        'reads': [],
        'breakdown': bd,
        'account': '2007', 'issue_date': issue_date,
        'start_date': bill_start_date,
        'finish_date': bill_finish_date,
        'mpan_core': '22 0003 0354 632',
        'reference': reference,
    }


class Parser():
    def __init__(self, f):
        self.book = open_workbook(file_contents=f.read())
        self.sheet = self.book.sheet_by_index(1)

        self.last_line = None
        lines = (self._set_last_line(i, l) for i, l in enumerate(f))
        self.reader = csv.reader(lines, skipinitialspace=True)
        self._line_number = None
        self._title_line = None

    @property
    def line_number(self):
        return None if self._line_number is None else self._line_number + 1

    def _set_last_line(self, i, line):
        self._line_numer = i
        self.last_line = line
        if i == 0:
            self._title_line = line
        return line

    def make_raw_bills(self):
        bills = []
        title_row = self.sheet.row(0)
        for row_index in range(1, self.sheet.nrows):
            row = self.sheet.row(row_index)
            datemode = self.book.datemode
            try:
                bills.append(_parse_row(row, row_index, datemode, title_row))
            except BadRequest as e:
                raise BadRequest(
                    "On row " + str(row_index) + ": " + str(e.description))
        return bills



from chellow.utils import get_file_rates


def hh(supply_source):
    try:
        supply_source.caches['ro']
    except KeyError:
        supply_source.caches['ro'] = {}

    for hh in supply_source.hh_data:
        rate = float(
            get_file_rates(
                supply_source.caches, 'ro', hh['start-date'])
            ['ro_gbp_per_msp_kwh'])
        hh['ro-rate'] = rate
        hh['ro-msp-kwh'] = hh['msp-kwh']
        hh['ro-gbp'] = hh['msp-kwh'] * rate



import traceback
from sqlalchemy import or_
from sqlalchemy.sql.expression import null
from sqlalchemy.orm import joinedload
from chellow.models import Session, GEra, GSupply
from chellow.utils import hh_format, req_date, req_int
import chellow.dloads
import sys
import os
import threading
from flask import g, request
import csv
from chellow.views import chellow_redirect


def content(running_name, finished_name, date, g_supply_id):
    sess = None
    try:
        sess = Session()
        f = open(running_name, mode='w', newline='')
        writer = csv.writer(f, lineterminator='\n')
        writer.writerow(
            (
                'Date', 'Physical Site Id', 'Physical Site Name',
                'Other Site Ids', 'Other Site Names', 'MPRN', 'Exit Zone',
                'Meter Serial Number', 'Correction Factor', 'Unit', 'Contract',
                'Account', 'Supply Start', 'Supply Finish'))

        g_eras = sess.query(GEra, GSupply).join(GSupply).filter(
            GEra.start_date <= date,
            or_(
                GEra.finish_date == null(),
                GEra.finish_date >= date)).order_by(
            GSupply.mprn).options(joinedload(GEra.site_g_eras))

        if g_supply_id is not None:
            g_supply = GSupply.get_by_id(sess, g_supply_id)

            g_eras = g_eras.filter(GEra.g_supply == g_supply)

        for g_era, g_supply in g_eras:
            site_codes = []
            site_names = []
            for site_g_era in g_era.site_g_eras:
                if site_g_era.is_physical:
                    physical_site = site_g_era.site
                else:
                    site = site_g_era.site
                    site_codes.append(site.code)
                    site_names.append(site.name)

            sup_g_eras = sess.query(GEra).filter(
                GEra.g_supply == g_supply).order_by(GEra.start_date).all()
            g_supply_start_date = sup_g_eras[0].start_date
            g_supply_finish_date = sup_g_eras[-1].finish_date

            writer.writerow(
                ('' if value is None else str(value)) for value in [
                    hh_format(date), physical_site.code, physical_site.name,
                    ', '.join(site_codes), ', '.join(site_names),
                    g_supply.mprn, g_supply.g_exit_zone.code, g_era.msn,
                    g_era.correction_factor, g_era.g_unit.code,
                    g_era.g_contract.name, g_era.account,
                    hh_format(g_supply_start_date),
                    hh_format(g_supply_finish_date, ongoing_str='')])
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        writer.writerow([msg])
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(session):
    user = g.user
    date = req_date('date')
    if 'g_supply_id' in request.values:
        g_supply_id = req_int('g_supply_id')
    else:
        g_supply_id = None

    running_name, finished_name = chellow.dloads.make_names(
        'g_supplies_snapshot.csv', user)

    args = (running_name, finished_name, date, g_supply_id)
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



import traceback
from sqlalchemy import or_
from sqlalchemy.sql.expression import null, true
from datetime import datetime
import pytz
from chellow.utils import prev_hh, hh_after, hh_before, req_int
from chellow.views import chellow_redirect
from chellow.models import Supply, Source, Era, Site, SiteEra, Session
from flask import request, g
import chellow.dloads
import csv
import sys
import os
import threading


def content(year, supply_id, user):
    sess = f = writer = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names(
            'crc_special_events.csv', user)
        f = open(running_name, mode='w', newline='')
        writer = csv.writer(f, lineterminator='\n')
        writer.writerow(("MPAN Core", "Site Id", "Site Name", "Date", "Event"))

        year_start = datetime(year, 4, 1, tzinfo=pytz.utc)
        year_finish = prev_hh(datetime(year + 1, 4, 1, tzinfo=pytz.utc))

        def add_event(events, date, code, era=None, mpan_core=None):
            if era is None:
                mpan_cores = [mpan_core]
            else:
                mpan_cores = []
                if era.imp_mpan_core is not None:
                    mpan_cores.append(era.imp_mpan_core)
                if era.exp_mpan_core is not None:
                    mpan_cores.append(era.exp_mpan_core)

            for mpan_core in mpan_cores:
                events.append(
                    {'date': date, 'code': code, 'mpan-core': mpan_core})

        if supply_id is None:
            supplies = sess.query(Supply).join(Source).join(Era).filter(
                Source.code.in_(('net', 'gen-net', 'gen')),
                Era.start_date <= year_finish, or_(
                    Era.finish_date == null(),
                    Era.finish_date >= year_start)).distinct()
        else:
            supply = Supply.get_by_id(supply_id)
            supplies = sess.query(Supply).filter(Supply.id == supply.id)

        for supply in supplies:
            eras = sess.query(Era).filter(
                Era.supply == supply, Era.start_date <= year_finish,
                or_(Era.finish_date == null(), Era.finish_date >= year_start)
                ).order_by(Era.start_date).all()
            events = []
            first_era = eras[0]
            first_era_start = first_era.start_date
            if hh_after(first_era_start, year_start):
                add_event(events, first_era_start, "New Supply", first_era)

            last_era = eras[-1]
            last_era_finish = last_era.finish_date
            if hh_before(last_era_finish, year_finish):
                add_event(events, last_era_finish, "Disconnection", last_era)

            prev_era = first_era
            for era in eras[1:]:
                if era.msn != prev_era.msn:
                    add_event(events, era.start_date, "Meter Change", era)
                if era.pc.code != prev_era.pc.code:
                    add_event(
                        events, era.start_date, "Change Of Profile Class", era)

                if era.mop_contract_id != prev_era.mop_contract_id:
                    add_event(events, era.start_date, "Change Of MOP", era)

                if era.dc_contract_id != prev_era.dc_contract_id:
                    add_event(events, era.start_date, "Change Of DC", era)

                for is_import in [True, False]:
                    if era.imp_mpan_core is None:
                        mpan_core = era.exp_mpan_core
                    else:
                        mpan_core = era.imp_mpan_core

                    if is_import:
                        cur_sup = era.imp_supplier_contract
                        prev_sup = prev_era.imp_supplier_contract
                    else:
                        cur_sup = era.exp_supplier_contract
                        prev_sup = prev_era.exp_supplier_contract

                    if cur_sup is None and prev_sup is not None:
                        add_event(
                            events, era.start_date, "End of supply", mpan_core)
                    elif cur_sup is not None and prev_sup is None:
                        add_event(
                            events, era.start_date, "Start of supply", None,
                            mpan_core)
                    elif cur_sup is not None and \
                            prev_sup is not None and cur_sup != prev_sup:
                        add_event(
                            events, era.start_date, "Change Of Supplier", None,
                            mpan_core)

                prev_era = era

            if len(events) > 0:
                site = sess.query(Site).join(SiteEra).filter(
                    SiteEra.is_physical == true(),
                    SiteEra.era == last_era).one()

                for event in events:
                    vals = [
                        event['mpan-core'], site.code, site.name,
                        event['date'].strftime("%Y-%m-%d %H:%M"),
                        event['code']]
                    writer.writerow(vals)

            # Avoid a long-running transaction
            sess.rollback()
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        writer.writerow([msg])
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    year = req_int('year')
    supply_id = req_int('supply_id') if 'supply_id' in request.values else None

    threading.Thread(target=content, args=(year, supply_id, g.user)).start()
    return chellow_redirect("/downloads", 303)



from datetime import datetime as Datetime
import datetime
import pytz
from sqlalchemy.sql.expression import null
from sqlalchemy.orm import joinedload
import traceback
from chellow.models import Snag, Site, Session
from chellow.views import chellow_redirect
import chellow.dloads
import csv
from flask import g
import threading
import sys
import os
from chellow.utils import hh_format


def content(user):
    sess = f = writer = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names(
            'site_snags.csv', user)
        f = open(running_name, mode='w', newline='')
        writer = csv.writer(f, lineterminator='\n')
        writer.writerow(
            (
                'Chellow Id', 'Site Code', 'Site Name', 'Snag Description',
                'Start Date', 'Finish Date', 'Days Since Snag Finished',
                'Duration Of Snag (Days)', 'Is Ignored?'))

        now = Datetime.now(pytz.utc)

        for snag in sess.query(Snag).join(Site).filter(
                Snag.site != null()).order_by(
                Site.code, Snag.description, Snag.start_date,
                Snag.id).options(joinedload(Snag.site)):
            snag_start = snag.start_date
            snag_finish = snag.finish_date
            if snag_finish is None:
                duration = now - snag_start
                age_of_snag = datetime.timedelta(0)
            else:
                duration = snag_finish - snag_start
                age_of_snag = now - snag_finish

            writer.writerow(
                (
                    str(snag.id), snag.site.code, snag.site.name,
                    snag.description, hh_format(snag_start),
                    hh_format(snag_finish),
                    str(age_of_snag.days + age_of_snag.seconds / (3600 * 24)),
                    str(duration.days + duration.seconds / (3600 * 24)),
                    str(snag.is_ignored)))
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        writer.writerow([msg])
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    args = (g.user,)
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



from sqlalchemy import or_
from sqlalchemy.sql.expression import null, true
import traceback
from chellow.models import Session, GSupply, GEra, Site, SiteGEra
from chellow.utils import (
    hh_min, hh_max, hh_format, req_int, req_date, csv_make_val)
from chellow.views import chellow_redirect
from chellow.computer import forecast_date
from werkzeug.exceptions import BadRequest
from chellow.g_engine import GDataSource, g_contract_func
import csv
import os
from flask import g
import threading
import chellow.dloads


def content(g_supply_id, file_name, start_date, finish_date, user):
    caches = {}
    sess = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names(
            'g_supply_virtual_bill_' + str(g_supply_id) + '.csv', user)
        f = open(running_name, mode='w', newline='')
        writer = csv.writer(f, lineterminator='\n')

        g_supply = GSupply.get_by_id(sess, g_supply_id)

        forecast_dt = forecast_date()

        prev_titles = None

        for g_era in sess.query(GEra).filter(
                GEra.g_supply == g_supply, GEra.start_date < finish_date, or_(
                    GEra.finish_date == null(),
                    GEra.finish_date > start_date)).order_by(GEra.start_date):

            chunk_start = hh_max(g_era.start_date, start_date)
            chunk_finish = hh_min(g_era.finish_date, finish_date)
            site = sess.query(Site).join(SiteGEra).filter(
                SiteGEra.g_era == g_era, SiteGEra.is_physical == true()).one()

            ds = GDataSource(
                sess, chunk_start, chunk_finish, forecast_dt, g_era, caches,
                None)

            titles = [
                'MPRN', 'Site Code', 'Site Name', 'Account', 'From', 'To', '']

            output_line = [
                g_supply.mprn, site.code, site.name, ds.account,
                hh_format(ds.start_date), hh_format(ds.finish_date), '']

            contract_titles = g_contract_func(
                caches, g_era.g_contract, 'virtual_bill_titles')()
            titles.extend(contract_titles)

            g_contract_func(caches, g_era.g_contract, 'virtual_bill')(ds)
            bill = ds.bill

            for title in contract_titles:
                if title in bill:
                    output_line.append(csv_make_val(bill[title]))
                    del bill[title]
                else:
                    output_line.append('')

            for k in sorted(bill.keys()):
                output_line.extend([k, bill[k]])

            if titles != prev_titles:
                prev_titles = titles
                writer.writerow([str(v) for v in titles])
            writer.writerow(output_line)
    except BadRequest as e:
        writer.writerow(["Problem: " + e.description])
    except BaseException:
        writer.writerow([traceback.format_exc()])
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    g_supply_id = req_int('g_supply_id')
    file_name = 'g_supply_virtual_bill_' + str(g_supply_id) + '.csv'
    start_date = req_date('start')
    finish_date = req_date('finish')
    args = (g_supply_id, file_name, start_date, finish_date, g.user)

    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



import zipfile
import traceback
from sqlalchemy.sql.expression import true, null
from sqlalchemy import or_
from sqlalchemy.orm import joinedload
from chellow.models import Site, SiteEra, Era, Session, Supply
from chellow.utils import (
    hh_format, req_date, req_int, to_utc, ct_datetime, to_ct)
from flask import request, g
import chellow.dloads
import sys
import os
import csv
from chellow.views import chellow_redirect
import threading
from io import StringIO


def none_content(site_id, start_date, finish_date, user, file_name):
    sess = zf = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names(
            file_name, user)
        sites = sess.query(Site).join(SiteEra).join(Era).filter(
            SiteEra.is_physical == true(), or_(
                Era.finish_date == null(), Era.finish_date >= start_date),
            Era.start_date <= finish_date)
        zf = zipfile.ZipFile(running_name, 'w')

        start_date_str = hh_format(start_date)
        finish_date_str = hh_format(finish_date)
        for site in sites:
            buf = StringIO()
            writer = csv.writer(buf, lineterminator='\n')
            writer.writerow(
                [
                    "Site Code", "Site Name", "Associated Site Codes",
                    "Sources", "Generator Types", "From", "To", "Type",
                    "Date"] + list(map(str, range(1, 51))))
            associates = ' '.join(
                s.code for s in site.find_linked_sites(
                    sess, start_date, finish_date))
            source_codes = set()
            gen_types = set()
            for supply in sess.query(Supply).join(Era).join(SiteEra).filter(
                    SiteEra.is_physical == true(), SiteEra.site == site,
                    Era.start_date <= finish_date, or_(
                        Era.finish_date == null(),
                        Era.finish_date >= start_date)).distinct().options(
                            joinedload(Supply.source),
                            joinedload(Supply.generator_type)):
                source_codes.add(supply.source.code)
                gen_type = supply.generator_type
                if gen_type is not None:
                    gen_types.add(gen_type.code)
            source_codes_str = ', '.join(sorted(source_codes))
            gen_types_str = ', '.join(sorted(gen_types))
            row = None
            for hh in site.hh_data(sess, start_date, finish_date):
                ct_start_date = to_ct(hh['start_date'])
                if ct_start_date.hour == 0 and ct_start_date.minute == 0:
                    if row is not None:
                        writer.writerow(row)
                    row = [
                        site.code, site.name, associates, source_codes_str,
                        gen_types_str, start_date_str, finish_date_str,
                        'used', ct_start_date.strftime('%Y-%m-%d')]
                used_gen_kwh = hh['imp_gen'] - hh['exp_net'] - hh['exp_gen']
                used_3p_kwh = hh['imp_3p'] - hh['exp_3p']
                used_kwh = hh['imp_net'] + used_gen_kwh + used_3p_kwh
                row.append(str(round(used_kwh, 2)))
            if row is not None:
                writer.writerow(row)
            zf.writestr(
                site.code + '_' +
                finish_date.strftime('%Y%m%d%M%H') + '.csv',
                buf.getvalue())

            # Avoid long-running transaction
            sess.rollback()
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        zf.write(msg)
    finally:
        if sess is not None:
            sess.close()
        if zf is not None:
            zf.close()
            os.rename(running_name, finished_name)


def site_content(site_id, start_date, finish_date, user, file_name):
    sess = f = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names(
            file_name, user)
        f = open(running_name, mode='w', newline='')
        writer = csv.writer(f, lineterminator='\n')
        site = Site.get_by_id(sess, site_id)
        sites = sess.query(Site).filter(Site.id == site_id)
        start_date_str = hh_format(start_date)
        finish_date_str = hh_format(finish_date)

        for site in sites:
            writer.writerow(
                [
                    "Site Code", "Site Name", "Associated Site Codes",
                    "Sources", "Generator Types", "From", "To", "Type",
                    "Date"] + list(map(str, range(1, 51))))
            associates = ' '.join(
                s.code for s in site.find_linked_sites(
                    sess, start_date, finish_date))
            source_codes = set()
            gen_types = set()
            for supply in sess.query(Supply).join(Era).join(SiteEra).filter(
                    SiteEra.is_physical == true(), SiteEra.site == site,
                    Era.start_date <= finish_date, or_(
                        Era.finish_date == null(),
                        Era.finish_date >= start_date)).distinct().options(
                            joinedload(Supply.source),
                            joinedload(Supply.generator_type)):
                source_codes.add(supply.source.code)
                gen_type = supply.generator_type
                if gen_type is not None:
                    gen_types.add(gen_type.code)
            source_codes_str = ', '.join(sorted(source_codes))
            gen_types_str = ', '.join(sorted(gen_types))
            vals = None
            for hh in site.hh_data(sess, start_date, finish_date):
                hh_start_ct = to_ct(hh['start_date'])
                if hh_start_ct.hour == 0 and hh_start_ct.minute == 0:
                    if vals is not None:
                        writer.writerow(vals)
                    vals = [
                        site.code, site.name, associates, source_codes_str,
                        gen_types_str, start_date_str, finish_date_str, 'used',
                        hh_start_ct.strftime('%Y-%m-%d')]
                used_gen_kwh = hh['imp_gen'] - hh['exp_net'] - hh['exp_gen']
                used_3p_kwh = hh['imp_3p'] - hh['exp_3p']
                used_kwh = hh['imp_net'] + used_gen_kwh + used_3p_kwh
                vals.append(str(round(used_kwh, 2)))
            if vals is not None:
                writer.writerow(vals)
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        f.write(msg)
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    start_date = req_date('start', 'day')
    finish_year = req_int('finish_year')
    finish_month = req_int('finish_month')
    finish_day = req_int('finish_day')
    finish_date_ct = ct_datetime(finish_year, finish_month, finish_day, 23, 30)
    finish_date = to_utc(finish_date_ct)

    finish_date_str = finish_date_ct.strftime('%Y%m%d%H%M')
    if 'site_id' in request.values:
        site_id = req_int('site_id')
        file_name = "sites_hh_data_" + str(site_id) + "_" + finish_date_str + \
            ".csv"
    else:
        site_id = None
        file_name = "supplies_hh_data_" + finish_date_str + ".zip"

    content = none_content if site_id is None else site_content
    args = (site_id, start_date, finish_date, g.user, file_name)
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



from decimal import Decimal
from datetime import datetime as Datetime
import csv
from dateutil.relativedelta import relativedelta
from werkzeug.exceptions import BadRequest
from chellow.utils import validate_hh_start, HH, to_utc, parse_mpan_core
from io import StringIO


col_map = {
    'Gsp Data': 'sum-gsp-kwh',
    'Levy Exempt Energy Units': 'lec-kwh',
    'Levy Exempt Energy Price': 'lec-rate',
    'Levy Exempt Energy Charge': 'lec-gbp',
    'Meter Reading Charges Charge': 'meter-reading-gbp',
    'Availability Charges Units': 'duos-availability-kva',
    'Availability Charges Price': 'duos-availability-rate',
    'Availability Charges Charge': 'duos-availability-gbp',
    'Excess Availability Charges Units': 'duos-excess-availability-kva',
    'Excess Availability Charges Price': 'duos-excess-availability-rate',
    'Excess Availability Charges Charge': 'duos-excess-availability-gbp',
    'Fixed Charges Units': 'duos-fixed-days',
    'Fixed Charges Price': 'duos-fixed-rate',
    'Fixed Charges Charge': 'duos-fixed-gbp',
    'Rate 1 Units': 'duos-red-kwh',
    'Rate 1 Price': 'duos-red-rate',
    'Rate 1 Charge': 'duos-red-gbp',
    'Rate 2 Units': 'duos-amber-kwh',
    'Rate 2 Price': 'duos-amber-rate',
    'Rate 2 Charge': 'duos-amber-gbp',
    'Rate 3 Units': 'duos-green-kwh',
    'Rate 3 Price': 'duos-green-rate',
    'Rate 3 Charge': 'duos-green-gbp',
    'Reactive Power Charge Units': 'duos-reactive-kvarh',
    'Reactive Power Charge Price': 'duos-reactive-rate',
    'Reactive Power Charge Charge': 'duos-reactive-gbp',
    'BSUoS Actual Charge (Date) Charge': 'bsuos-actual-gbp-info',
    'BSUOSTTL': 'bsuos-gbp',
    'AAHEDC Charge Units': 'aahedc-gsp-kwh',
    'AAHEDC Charge Price': 'aahedc-rate',
    'AAHEDC Charge Charge': 'aahedc-gbp',
    'RO Adjustment Units': 'ro-msp-kwh',
    'RO Adjustment Price': 'ro-rate',
    'RO Adjustment Charge': 'ro-gbp',
    'Estimated FiT Charge Units': 'fit-estimate-msp-kwh',
    'Estimated FiT Charge Price': 'fit-estimate-rate',
    'Estimated FiT Charge Charge': 'fit-estimate-gbp',
    'Actual FiT Charge Price': 'fit-previous-actual-rate',
    'Actual FiT Charge Charge': 'fit-previous-actual-gbp',
    'Estimated CfD Charge Units': 'cfd-fit-estimate-nbp-kwh',
    'Estimated CfD Charge Price': 'cfd-fit-estimate-rate',
    'Estimated CfD Charge Charge': 'cfd-fit-estimate-gbp',
    'Actual CfD Charge Units': 'cfd-fit-previous-actual-nbp-kwh',
    'Actual CfD Charge Price': 'cfd-fit-previous-actual-rate',
    'Actual CfD Charge Charge': 'cfd-fit-previous-actual-gbp',
    'CFD_Total': 'cfd-fit-total',
    'AHC_NRG1': 'reconciliation-gbp',
    'Energy Peak Units': 'peak-gsp-kwh',
    'Energy Peak Price': 'peak-rate',
    'Energy Peak Charge': 'peak-gbp',
    'Energy Peak Shoulder Units': 'peak-shoulder-gsp-kwh',
    'Energy Peak Shoulder Price': 'peak-shoulder-rate',
    'Energy Peak Shoulder Charge': 'peak-shoulder-gbp',
    'Energy Summer Night Units': 'summer-night-gsp-kwh',
    'Energy Summer Night Price': 'summer-night-rate',
    'Energy Summer Night Charge': 'summer-night-gbp',
    'Energy Summer Weekday Units': 'summer-weekday-gsp-kwh',
    'Energy Summer Weekday Price': 'summer-weekday-rate',
    'Energy Summer Weekday Charge': 'summer-weekday-gbp',
    'Energy Summer Weekend Units': 'summer-weekend-gsp-kwh',
    'Energy Summer Weekend Price': 'summer-weekend-rate',
    'Energy Summer Weekend Charge': 'summer-weekend-gbp',
    'Energy Winter Night Units': 'winter-night-gsp-kwh',
    'Energy Winter Night Price': 'winter-night-rate',
    'Energy Winter Night Charge': 'winter-night-gbp',
    'Energy Winter Weekday Units': 'winter-weekday-gsp-kwh',
    'Energy Winter Weekday Price': 'winter-weekday-rate',
    'Energy Winter Weekday Charge': 'winter-weekday-gbp',
    'Energy Winter Weekend Units': 'winter-weekend-gsp-kwh',
    'Energy Winter Weekend Price': 'winter-weekend-rate',
    'Energy Winter Weekend Charge': 'winter-weekend-gbp',
    'Network UoS Charges Price': 'triad-estimate-rate',
    'Network UoS Charges Units': 'triad-estimate-gsp-kw',
    'Network UoS Charges Charge': 'triad-estimate-gbp',
    'CAP_M_TTL': 'capacity-market-gbp'
}


class Parser():
    def __init__(self, f):
        self.last_line = None
        tf = StringIO(str(f.read(), 'utf-8', errors='ignore'))
        lines = (self._set_last_line(i, l) for i, l in enumerate(tf))
        self.reader = csv.reader(lines, skipinitialspace=True)
        self._line_number = None
        self._title_line = None

    @property
    def line_number(self):
        return None if self._line_number is None else self._line_number + 1

    def _set_last_line(self, i, line):
        self._line_numer = i
        self.last_line = line
        if i == 0:
            self._title_line = line
        return line

    def make_raw_bills(self):
        titles = [t.strip() for t in next(iter(self.reader))]
        title_idx = dict((title, i) for i, title in enumerate(titles))

        raw_bills = []

        for vals in self.reader:
            if len(vals) == 0 or vals[0].startswith('#') or \
                    ''.join(vals) == '':
                continue

            def val(title):
                idx = title_idx[title]
                try:
                    return vals[idx]
                except KeyError:
                    raise BadRequest(
                        "For title " + title + " and index " + str(idx))

            def date_val(title):
                try:
                    return to_utc(Datetime.strptime(val(title), "%d/%m/%Y"))
                except ValueError as e:
                    raise BadRequest(
                        "At line number " + str(self._line_number) +
                        ", while trying to find the value of " + title +
                        " the date couldn't be parsed. " + str(e) +
                        " The full line is " + self.last_line)

            def dec_val(title):
                return Decimal(val(title))

            issue_date = date_val('Invoice Date')
            bill_from = validate_hh_start(date_val('Bill Period Start'))
            bill_to = validate_hh_start(date_val('Bill Period End')) + \
                relativedelta(days=1) - HH
            kwh = dec_val('Site Data')
            breakdown = dict(
                [(v, dec_val(k)) for k, v in col_map.items() if k in titles])
            breakdown['raw_lines'] = [
                self._title_line.strip(), self.last_line.strip()]
            net = Decimal('0.00') + dec_val('NET_AMT')
            vat = Decimal('0.00') + dec_val('VATTTL')
            gross = Decimal('0.00') + dec_val('TTLCHG')
            mpan_core = parse_mpan_core(val('Mpan'))
            raw_bill = {
                'bill_type_code': 'N', 'account': val('Bill Ref No.'),
                'mpan_core': mpan_core, 'reference': val('Invoice No.'),
                'issue_date': issue_date, 'start_date': bill_from,
                'finish_date': bill_to, 'kwh': kwh, 'net': net, 'vat': vat,
                'gross': gross, 'breakdown': breakdown, 'reads': []
            }
            raw_bills.append(raw_bill)

        return raw_bills



from datetime import datetime as Datetime, timedelta as Timedelta
import csv
import os
import io
from flask import render_template, request, g
from chellow.models import (
    VoltageLevel, Participant, Party, MarketRole, Llfc, Mtc, MeterType,
    Session)
from chellow.utils import hh_format, to_utc, to_ct, ct_datetime
from werkzeug.exceptions import BadRequest
from sqlalchemy.orm import joinedload
from sqlalchemy import null
import traceback
import chellow.dloads
import threading
from chellow.views import chellow_redirect


def parse_date(date_str):
    if len(date_str) == 0:
        return None
    else:
        return to_utc(to_ct(Datetime.strptime(date_str, "%d/%m/%Y")))


def parse_to_date(date_str):
    if len(date_str) == 0:
        return None
    else:
        dt = to_ct(Datetime.strptime(date_str, "%d/%m/%Y"))
        dt += Timedelta(hours=23, minutes=30)
        return to_utc(dt)


def is_common_mtc(code):
    return 499 < code < 510 or 799 < code < 1000


def do_get(sess):
    return render_template('report_163.html')


VOLTAGE_MAP = {
    '24': {
        '602': {
            to_utc(ct_datetime(2010, 4, 1)): 'LV'
        }
    }
}


def content(table, version, fin, user):
    sess = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names(
            table + '_' + version + '_general_import.csv', user)
        f = open(running_name, mode='w', newline='')
        w = csv.writer(f, lineterminator='\n')

        reader = iter(csv.reader(fin))
        next(reader)
        if table == 'Line_Loss_Factor_Class':
            VOLTAGE_LEVEL_CODES = set(
                [v.code for v in sess.query(VoltageLevel)])
            DNO_MAP = dict(
                (dno.participant.code, dno) for dno in sess.query(Party).
                join(MarketRole).filter(MarketRole.code == 'R').options(
                    joinedload(Party.participant)))
            for i, values in enumerate(reader):
                participant_code = values[0]
                # market_role_code = values[1]
                llfc_code = values[3].zfill(3)
                valid_from = parse_date(values[4])
                description = values[5]
                is_import = values[6] in ('A', 'B')
                is_substation = any(
                    p in description for p in (
                        '_SS', ' SS', ' S/S',  '(S/S)', 'sub', 'Sub'))

                valid_to = parse_to_date(values[7])

                try:
                    dno = DNO_MAP[participant_code]
                except KeyError:
                    w.writerow(
                        (
                            "# There is no DNO with participant code ",
                            participant_code))
                    continue

                try:
                    voltage_level_code = VOLTAGE_MAP[dno.dno_code][llfc_code][
                        valid_from]
                except KeyError:
                    voltage_level_code = 'LV'
                    description_upper = description.upper()
                    for vl_code in VOLTAGE_LEVEL_CODES:
                        if vl_code in description_upper:
                            voltage_level_code = vl_code
                            break

                llfc = sess.query(Llfc).filter(
                    Llfc.dno == dno, Llfc.code == llfc_code,
                    Llfc.valid_from == valid_from).first()
                if llfc is None:
                    w.writerow(
                        (
                            'insert', 'llfc', dno.dno_code, llfc_code,
                            description, voltage_level_code, is_substation,
                            is_import, hh_format(valid_from),
                            hh_format(valid_to, ongoing_str='')))
                elif any(
                        (
                            description != llfc.description,
                            voltage_level_code != llfc.voltage_level.code,
                            is_substation != llfc.is_substation,
                            is_import != llfc.is_import,
                            valid_to != llfc.valid_to)):
                    w.writerow(
                        (
                            'update', 'llfc', dno.dno_code, llfc.code,
                            hh_format(llfc.valid_from), description,
                            voltage_level_code, is_substation, is_import,
                            hh_format(valid_to, ongoing_str='')))

        elif table == 'Market_Participant':
            for i, values in enumerate(reader):
                participant_code = values[0]
                participant_name = values[1]

                participant = sess.query(Participant).filter(
                    Participant.code == participant_code).first()

                if participant is None:
                    w.writerow(
                        (
                            'insert', 'participant', participant_code,
                            participant_name))

                elif participant_name != participant.name:
                    w.writerow(
                        (
                            'update', 'participant', participant_code,
                            participant_name))

        elif table == 'Market_Role':
            for i, values in enumerate(reader):
                role_code = values[0]
                role_description = values[1]

                role = sess.query(MarketRole).filter(
                    MarketRole.code == role_code).first()

                if role is None:
                    w.writerow(
                        (
                            'insert', 'market_role', role_code,
                            role_description))

                elif role_description != role.description:
                    w.writerow(
                        (
                            'update', 'market_role', role_code,
                            role_description))

        elif table == 'Market_Participant_Role':
            for i, values in enumerate(reader):
                participant_code = values[0]
                market_role_code = values[1]
                valid_from = parse_date(values[2])
                party = sess.query(Party).join(Participant). \
                    join(MarketRole).filter(
                        Party.valid_from == valid_from,
                        Participant.code == participant_code,
                        MarketRole.code == market_role_code).first()
                valid_to = parse_to_date(values[3])
                name = values[4]
                dno_code_str = values[14]
                dno_code = None if len(dno_code_str) == 0 else dno_code_str
                if dno_code == '99':
                    continue

                if party is None:
                    w.writerow(
                        (
                            'insert', 'party', market_role_code,
                            participant_code, name, hh_format(valid_from),
                            hh_format(valid_to, ongoing_str=''),
                            dno_code_str))
                elif any(
                        (
                            name != party.name, dno_code != party.dno_code,
                            valid_to != party.valid_to)):
                    w.writerow(
                        (
                            'update', 'party', market_role_code,
                            participant_code, name, hh_format(valid_from),
                            hh_format(valid_to, ongoing_str=''),
                            dno_code_str))

        elif table == 'Meter_Timeswitch_Class':
            for i, values in enumerate(reader):
                code = values[0].zfill(3)
                valid_from = parse_date(values[1])
                valid_to = parse_to_date(values[2])
                description = values[3]
                is_common = values[4] == 'T'
                has_related_metering = values[5] == 'T'
                meter_type_code = values[6]
                meter_payment_type_code = values[7]
                has_comms = values[8] == 'T'
                is_hh = values[9] == 'H'
                tpr_count_str = values[10]
                tpr_count = 0 if tpr_count_str == '' else int(tpr_count_str)

                if is_common:
                    mtc = sess.query(Mtc).filter(
                        Mtc.dno == null(), Mtc.code == code,
                        Mtc.valid_from == valid_from).first()
                    if mtc is None:
                        w.writerow(
                            (
                                'insert', 'mtc', '', code, description,
                                has_related_metering, has_comms, is_hh,
                                meter_type_code, meter_payment_type_code,
                                tpr_count, hh_format(valid_from),
                                hh_format(valid_to, ongoing_str='')))
                    elif any(
                            (
                                description != mtc.description,
                                has_related_metering !=
                                mtc.has_related_metering,
                                has_comms != mtc.has_comms,
                                is_hh != mtc.is_hh,
                                meter_type_code != mtc.meter_type.code,
                                meter_payment_type_code !=
                                mtc.meter_payment_type.code,
                                tpr_count != mtc.tpr_count,
                                valid_to != mtc.valid_to)):
                        w.writerow(
                            (
                                'update', 'mtc', '', mtc.code, description,
                                has_related_metering, has_comms, is_hh,
                                meter_type_code, meter_payment_type_code,
                                tpr_count, hh_format(mtc.valid_from),
                                hh_format(valid_to, ongoing_str='')))

        elif table == 'MTC_in_PES_Area':
            dnos = dict(
                (p.participant.code, (p.id, p.dno_code)) for p in sess.query(
                    Party).join(Participant).join(MarketRole).filter(
                    MarketRole.code == 'R').options(
                    joinedload(Party.participant)))
            mtcs = dict(
                ((m.dno_id, m.code, m.valid_from), m)
                for m in sess.query(Mtc).options(
                    joinedload(Mtc.meter_type),
                    joinedload(Mtc.meter_payment_type)).all())
            for i, values in enumerate(reader):
                code_str = values[0]
                if not Mtc.has_dno(code_str):
                    continue

                code_int = int(code_str)
                code = code_str.zfill(3)
                participant_code = values[2]
                dno_id, dno_code = dnos[participant_code]
                valid_from = parse_date(values[3])
                valid_to = parse_to_date(values[4])
                description = values[5]
                meter_type_code = values[6]
                meter_payment_type_code = values[7]
                has_related_metering = code_int > 500
                has_comms = values[8] == 'Y'
                is_hh = values[9] == 'H'
                tpr_count_str = values[10]
                tpr_count = 0 if tpr_count_str == '' else int(tpr_count_str)

                mtc = mtcs.get((dno_id, code, valid_from))

                if mtc is None:
                    w.writerow(
                        (
                            'insert', 'mtc', dno_code, code, description,
                            has_related_metering, has_comms, is_hh,
                            meter_type_code, meter_payment_type_code,
                            tpr_count, hh_format(valid_from),
                            hh_format(valid_to, ongoing_str='')))
                elif any(
                        (
                            description != mtc.description,
                            has_related_metering != mtc.has_related_metering,
                            has_comms != mtc.has_comms,
                            is_hh != mtc.is_hh,
                            meter_type_code != mtc.meter_type.code,
                            meter_payment_type_code !=
                            mtc.meter_payment_type.code,
                            tpr_count != mtc.tpr_count,
                            valid_to != mtc.valid_to)):
                    w.writerow(
                        (
                            'update', 'mtc', mtc.dno.dno_code, mtc.code,
                            description, has_related_metering, has_comms,
                            is_hh, meter_type_code, meter_payment_type_code,
                            tpr_count, hh_format(mtc.valid_from),
                            hh_format(valid_to, ongoing_str='')))

        elif table == 'MTC_Meter_Type':
            for i, values in enumerate(reader):
                code = values[0]
                description = values[1]
                valid_from = parse_date(values[2])
                valid_to = parse_to_date(values[3])
                pt = sess.query(MeterType).filter(
                    MeterType.code == code,
                    MeterType.valid_from == valid_from).first()
                if pt is None:
                    w.writerow(
                        (
                            'insert', 'meter_type', code, description,
                            hh_format(valid_from),
                            hh_format(valid_to, ongoing_str='')))

                elif (description, valid_from, valid_to) != (
                        pt.description, pt.valid_from, pt.valid_to):
                    w.writerow(
                        (
                            'update', 'meter_type', code, description,
                            hh_format(valid_from), hh_format(valid_to)))
        else:
            raise Exception("The table " + table + " is not recognized.")
    except BaseException:
        w.writerow([traceback.format_exc()])
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_post(sess):
    file_item = request.files["file"]
    file_path = file_item.filename
    file_head, file_name = os.path.split(file_path)
    file_title, file_ext = os.path.splitext(file_name)
    if not file_ext == '.csv':
        raise BadRequest(
            "The file name should have the extension .csv, but in fact it "
            "has the extension '" + file_ext + "'.")
    idx = file_title.rfind('_')
    table = file_title[:idx]
    version = file_title[idx+1:]
    f = io.StringIO(str(file_item.read(), 'utf8'))
    f.seek(0)

    args = (table, version, f, g.user)
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



import csv
from decimal import Decimal, InvalidOperation
from io import StringIO
from itertools import count

from chellow.utils import parse_hh_start

from werkzeug.exceptions import BadRequest

from zish import ZishLocationException, loads


def get_str(row, idx, name, lineno):
    return row[idx].strip()


def get_datetime(row, idx, name, lineno):
    val = get_str(row, idx, name, lineno)
    try:
        return parse_hh_start(val)
    except BadRequest as e:
        raise BadRequest(
            "Problem parsing the timestamp '" + val + "' at line number " +
            str(lineno) + " for the field '" + name + "' in column " +
            str(idx + 1) + ": " + e.description)


def get_decimal(row, idx, name, lineno):
    val = get_str(row, idx, name, lineno)
    try:
        return Decimal(val)
    except InvalidOperation as e:
        raise BadRequest(
            "Problem parsing the decimal '" + val + "' at line number " +
            str(lineno) + " for the field '" + name + "' at " + str(idx + 1) +
            ": " + str(e))


class Parser():
    def __init__(self, file_bytes):
        self.csv_reader = csv.reader(
            StringIO(file_bytes.decode('utf8', errors='ignore')),
            skipinitialspace=True)
        self.titles = ','.join(next(self.csv_reader))
        self._line_number = None

    @property
    def line_number(self):
        return None if self._line_number is None else self._line_number + 2

    def _set_last_line(self, line):
        self.last_line = line
        return line

    def make_raw_bills(self):
        raw_bills = []
        for self._line_number, row in enumerate(self.csv_reader):

            bill_reference = row[0]
            if bill_reference == '' or bill_reference.startswith('#'):
                continue

            mprn = row[1]
            bill_type = row[2]
            account = row[3]
            issue_date = get_datetime(row, 4, 'Issue Date', self.line_number)
            start_date = get_datetime(row, 5, 'Start Date', self.line_number)
            finish_date = get_datetime(row, 6, 'Finish Date', self.line_number)
            kwh = get_decimal(row, 7, 'kWh', self.line_number)
            net = get_decimal(row, 8, 'Net GBP', self.line_number)
            vat = get_decimal(row, 9, 'VAT GBP', self.line_number)
            gross = get_decimal(row, 10, 'Gross GBP', self.line_number)
            try:
                breakdown = loads(row[11])
            except ZishLocationException as e:
                raise BadRequest(
                    "Problem parsing the breakdown field at line number " +
                    str(self.line_number) + ": " + str(e))

            reads = []

            for i in count(12, 10):
                if i > len(row) - 1 or len(''.join(row[i:]).strip()) == 0:
                    break

                msn = get_str(row, i, 'Meter Serial Number', self.line_number)
                unit = get_str(row, i + 1, 'Unit', self.line_number).upper()
                correction_factor = get_decimal(
                    row, i + 2, 'Correction Factor', self.line_number)

                calorific_value = get_str(
                    row, i + 3, 'Calorific Value', self.line_number)
                if len(calorific_value) > 0:
                    calorific_value = get_decimal(
                        row, i + 3, 'Calorific Value', self.line_number)
                else:
                    calorific_value = None

                prev_date = get_datetime(
                    row, i + 4, 'Previous Date', self.line_number)
                prev_value = get_decimal(
                    row, i + 5, 'Previous Value', self.line_number)
                prev_type = get_str(
                    row, i + 6, 'Previous Type', self.line_number)
                pres_date = get_datetime(
                    row, i + 7, 'Previous Date', self.line_number)
                pres_value = get_decimal(
                    row, i + 8, 'Present Value', self.line_number)
                pres_type = get_str(
                    row, i + 9, 'Present Type', self.line_number)
                reads.append(
                    {
                        'msn': msn,
                        'unit': unit,
                        'correction_factor': correction_factor,
                        'calorific_value': calorific_value,
                        'prev_date': prev_date,
                        'prev_value': prev_value,
                        'prev_type_code': prev_type,
                        'pres_date': pres_date,
                        'pres_value': pres_value,
                        'pres_type_code': pres_type})

            raw_bills.append(
                {
                    'mprn': mprn, 'reference': bill_reference,
                    'account': account, 'reads': reads, 'kwh': kwh,
                    'breakdown': breakdown, 'net_gbp': net, 'vat_gbp': vat,
                    'gross_gbp': gross,
                    'raw_lines': self.titles + '\n' + ','.join(row),
                    'bill_type_code': bill_type, 'start_date': start_date,
                    'finish_date': finish_date, 'issue_date': issue_date
                })
        return raw_bills



import traceback
from sqlalchemy import or_
from sqlalchemy.sql.expression import null, true
from chellow.models import Supply, Era, Session, Site, SiteEra
from chellow.utils import (
    req_date, hh_format, req_str, req_int, parse_mpan_core, req_bool, hh_range,
    to_ct)
import zipfile
import threading
import os
import chellow.dloads
from flask import request, g
from werkzeug.exceptions import BadRequest
from chellow.views import chellow_redirect
from datetime import datetime as Datetime


def csv_str(row):
    frow = []
    for cell in row:
        if cell is None:
            val = ''
        elif isinstance(cell, Datetime):
            val = hh_format(cell)
        else:
            val = str(cell)
        frow.append(val)

    return ','.join('"' + v + '"' for v in frow) + '\n'


def content(start_date, finish_date, supply_id, mpan_cores, is_zipped, user):
    if is_zipped:
        file_extension = ".zip"
    else:
        file_extension = ".csv"

    base_name = "hh_data_row_" + to_ct(start_date).strftime("%Y%m%d%H%M") + \
        file_extension

    tls = [
        "Site Code", "Imp MPAN Core", "Exp Mpan Core", "HH Start Clock-Time"]
    for polarity in ('Import', 'Export'):
        for suffix in (
                "ACTIVE kWh", "ACTIVE Status", "ACTIVE Modified",
                "REACTIVE_IMP kVArh", "REACTIVE_IMP Status",
                "REACTIVE_IMP Modified", "REACTIVE_EXP kVArh",
                "REACTIVE_EXP Status", "REACTIVE_EXP Modified"):
            tls.append(polarity + ' ' + suffix)

    titles = csv_str(tls)

    running_name, finished_name = chellow.dloads.make_names(base_name, user)

    if is_zipped:
        zf = zipfile.ZipFile(running_name, 'w')
    else:
        tmp_file = open(running_name, "w")
    sess = None
    try:
        sess = Session()
        caches = {}
        supplies = sess.query(Supply).join(Era).filter(
            Era.start_date <= finish_date, or_(
                Era.finish_date == null(),
                Era.finish_date >= start_date)).order_by(
            Era.supply_id, Era.start_date).distinct()
        if supply_id is not None:
            sup = Supply.get_by_id(sess, supply_id)
            supplies = supplies.filter(Era.supply == sup)

        if mpan_cores is not None:
            supplies = supplies.filter(
                or_(
                    Era.imp_mpan_core.in_(mpan_cores),
                    Era.exp_mpan_core.in_(mpan_cores)))

        if not is_zipped:
            tmp_file.write(titles)

        for supply in supplies:
            site, era = sess.query(
                Site, Era).join(Era.site_eras).filter(
                Era.supply == supply, Era.start_date <= finish_date,
                SiteEra.site_id == Site.id, or_(
                    Era.finish_date == null(),
                    Era.finish_date >= start_date),
                SiteEra.is_physical == true()).order_by(Era.id).first()

            outs = []
            data = iter(
                sess.execute("""
select hh_base.start_date,
    max(imp_active.value), max(imp_active.status),
    max(imp_active.last_modified),
    max(imp_reactive_imp.value), max(imp_reactive_imp.status),
    max(imp_reactive_imp.last_modified),
    max(imp_reactive_exp.value), max(imp_reactive_exp.status),
    max(imp_reactive_exp.last_modified),
    max(exp_active.value), max(exp_active.status),
    max(exp_active.last_modified),
    max(exp_reactive_imp.value), max(imp_reactive_imp.status),
    max(exp_reactive_imp.last_modified),
    max(exp_reactive_exp.value), max(exp_reactive_exp.status),
    max(exp_reactive_exp.last_modified)
from hh_datum hh_base
    join channel on hh_base.channel_id = channel.id
    join era on channel.era_id = era.id
    left join hh_datum imp_active
        on (imp_active.id = hh_base.id and channel.imp_related is true and
            channel.channel_type = 'ACTIVE')
    left join hh_datum imp_reactive_imp
        on (imp_reactive_imp.id = hh_base.id
            and channel.imp_related is true and
            channel.channel_type = 'REACTIVE_IMP')
    left join hh_datum imp_reactive_exp
        on (imp_reactive_exp.id = hh_base.id
            and channel.imp_related is true and
            channel.channel_type = 'REACTIVE_EXP')
    left join hh_datum exp_active
        on (exp_active.id = hh_base.id and channel.imp_related is false and
            channel.channel_type = 'ACTIVE')
    left join hh_datum exp_reactive_imp
                on (exp_reactive_imp.id = hh_base.id
                    and channel.imp_related is
                false and channel.channel_type = 'REACTIVE_IMP')
    left join hh_datum exp_reactive_exp
                on (exp_reactive_exp.id = hh_base.id
                and channel.imp_related is false
                and channel.channel_type = 'REACTIVE_EXP')
where supply_id = :supply_id
    and hh_base.start_date between :start_date and :finish_date
group by hh_base.start_date
order by hh_base.start_date
    """, params={
                    'supply_id': supply.id, 'start_date': start_date,
                    'finish_date': finish_date}))
            datum = next(data, None)

            for dt in hh_range(caches, start_date, finish_date):
                row = [site.code, era.imp_mpan_core, era.exp_mpan_core, dt]
                if datum is not None:
                    (
                        hh_start_date, imp_active, imp_active_status,
                        imp_active_modified, imp_reactive_imp,
                        imp_reactive_imp_status, imp_reactive_imp_modified,
                        imp_reactive_exp, imp_reactive_exp_status,
                        imp_reactive_exp_modified, exp_active,
                        exp_active_status, exp_active_modified,
                        exp_reactive_imp, exp_reactive_imp_status,
                        exp_reactive_imp_modified, exp_reactive_exp,
                        exp_reactive_exp_status, exp_reactive_exp_modified
                    ) = datum
                    if hh_start_date == dt:
                        datum = next(data, None)
                        row += [
                            imp_active, imp_active_status, imp_active_modified,
                            imp_reactive_imp, imp_reactive_imp_status,
                            imp_reactive_imp_modified, imp_reactive_exp,
                            imp_reactive_exp_status, imp_reactive_exp_modified,
                            exp_active, exp_active_status, exp_active_modified,
                            exp_reactive_imp, exp_reactive_imp_status,
                            exp_reactive_imp_modified, exp_reactive_exp,
                            exp_reactive_exp_status, exp_reactive_exp_modified]

                outs.append(csv_str(row))

            if is_zipped:
                zf.writestr(
                    (
                        "hh_data_row_" + str(era.id) + "_" +
                        str(era.imp_mpan_core) + "_" +
                        str(era.exp_mpan_core)).replace(' ', '') + '.csv',
                    titles + ''.join(outs))
            else:
                tmp_file.write(''.join(outs))

            # Avoid a long-running transaction
            sess.rollback()
    except BaseException:
        msg = "Problem " + traceback.format_exc()
        if is_zipped:
            zf.writestr('error.txt', msg)
        else:
            tmp_file.write(msg)
    finally:
        if sess is not None:
            sess.close()
        if is_zipped:
            zf.close()
        else:
            tmp_file.close()
        os.rename(running_name, finished_name)


def do_post(sess):
    start_date = req_date('start')
    finish_date = req_date('finish')
    supply_id = req_int('supply_id') if 'supply_id' in request.values else None

    if 'mpan_cores' in request.values:
        mpan_cores_str = req_str('mpan_cores')
        mpan_cores = mpan_cores_str.splitlines()
        if len(mpan_cores) == 0:
            mpan_cores = None
        else:
            for i in range(len(mpan_cores)):
                mpan_cores[i] = parse_mpan_core(mpan_cores[i])
    else:
        mpan_cores = None

    if finish_date < start_date:
        raise BadRequest("The finish date can't be before the start date.")

    is_zipped = req_bool('is_zipped')
    user = g.user

    threading.Thread(
        target=content, args=(
            start_date, finish_date, supply_id, mpan_cores, is_zipped, user)
        ).start()
    return chellow_redirect("/downloads", 303)



import atexit
import collections
import threading
import traceback
from datetime import datetime as Datetime
from decimal import Decimal

from chellow.models import (
    Contract, RateScript, Session, get_non_core_contract_id,
)
from chellow.utils import HH, hh_format, to_ct, to_utc, utc_datetime_now

from dateutil.relativedelta import relativedelta

import requests

from sqlalchemy import or_
from sqlalchemy.sql.expression import null

from werkzeug.exceptions import BadRequest

import xlrd

from zish import dumps, loads


def hh(data_source, run='RF'):
    try:
        bsuos_cache = data_source.caches['bsuos'][run]
    except KeyError:
        try:
            top_cache = data_source.caches['bsuos']
        except KeyError:
            top_cache = data_source.caches['bsuos'] = {}

        try:
            bsuos_cache = top_cache[run]
        except KeyError:
            bsuos_cache = top_cache[run] = {}

    for h in data_source.hh_data:
        try:
            h['bsuos-rate'] = bsuos_rate = bsuos_cache[h['start-date']]
        except KeyError:
            h_start = h['start-date']
            db_id = get_non_core_contract_id('bsuos')

            rates = data_source.hh_rate(db_id, h_start)['rates_gbp_per_mwh']

            try:
                bsuos_price_dict = rates[key_format(h_start)]
                bsuos_price = _find_price(run, bsuos_price_dict)
            except KeyError:
                ds = rates.values()
                bsuos_price = sum(_find_price(run, d) for d in ds) / len(ds)

            h['bsuos-rate'] = bsuos_rate = bsuos_cache[h_start] = float(
                bsuos_price) / 1000

        h['bsuos-kwh'] = h['nbp-kwh']
        h['bsuos-gbp'] = h['nbp-kwh'] * bsuos_rate


def _find_price(run, prices):
    try:
        price = prices[run]
    except KeyError:
        try:
            price = prices['RF']
        except KeyError:
            try:
                price = prices['SF']
            except KeyError:
                price = prices['II']
    return price


def key_format(dt):
    return dt.strftime("%d %H:%M")


bsuos_importer = None


def _save_cache(sess, cache):
    for yr, yr_cache in cache.items():
        for month, (rs, rates, rts) in tuple(yr_cache.items()):
            rs.script = dumps(rates)
            sess.commit()
            del yr_cache[month]


class BsuosImporter(threading.Thread):
    def __init__(self):
        super(BsuosImporter, self).__init__(name="BSUoS Importer")
        self.lock = threading.RLock()
        self.messages = collections.deque(maxlen=500)
        self.stopped = threading.Event()
        self.going = threading.Event()
        self.PROXY_HOST_KEY = 'proxy.host'
        self.PROXY_PORT_KEY = 'proxy.port'

    def stop(self):
        self.stopped.set()
        self.going.set()
        self.join()

    def go(self):
        self.going.set()

    def is_locked(self):
        if self.lock.acquire(False):
            self.lock.release()
            return False
        else:
            return True

    def log(self, message):
        self.messages.appendleft(
            utc_datetime_now().strftime("%Y-%m-%d %H:%M:%S") + " - " + message)

    def run(self):
        while not self.stopped.isSet():
            if self.lock.acquire(False):
                sess = self.global_alert = None
                try:
                    sess = Session()
                    self.log("Starting to check BSUoS rates.")
                    contract = Contract.get_non_core_by_name(sess, 'bsuos')
                    props = contract.make_properties()
                    if props.get('enabled', False):
                        urls = set(props.get('urls', []))
                        if props.get('discover_urls', False):
                            urls.update(_discover_urls(self.log))

                        url_list = sorted(urls)
                        self.log(
                            "List of URLs to process: " + str(url_list))
                        for url in url_list:
                            self.process_url(sess, url, contract)
                    else:
                        self.log(
                            "The automatic importer is disabled. To "
                            "enable it, edit the contract properties to "
                            "set 'enabled' to True.")
                except BaseException:
                    self.log("Outer problem " + traceback.format_exc())
                    self.global_alert = \
                        "There's a problem with the BSUoS automatic importer."
                    sess.rollback()
                finally:
                    if sess is not None:
                        sess.close()
                    self.lock.release()
                    self.log("Finished checking BSUoS rates.")

            self.going.wait(60 * 60 * 24)
            self.going.clear()

    def process_url(self, sess, url, contract):
        self.log("Checking to see if there's any new data at " + url)
        res = requests.get(url)
        self.log("Received " + str(res.status_code) + " " + res.reason)
        book = xlrd.open_workbook(file_contents=res.content)
        sheet = book.sheet_by_index(0)
        cache = {}

        for row_index in range(1, sheet.nrows):
            row = sheet.row(row_index)

            raw_date_val = row[0].value
            if isinstance(raw_date_val, float):
                raw_date = Datetime(
                    *xlrd.xldate_as_tuple(raw_date_val, book.datemode))
            elif isinstance(raw_date_val, str):
                separator = raw_date_val[2]
                fmat = separator.join(("%d", "%m", "%Y"))
                raw_date = Datetime.strptime(raw_date_val, fmat)
            else:
                raise BadRequest(
                    "Type of date field " + str(raw_date_val) +
                    " not recognized.")

            hh_date_ct = to_ct(raw_date)
            hh_date_ct += relativedelta(minutes=30*(int(row[1].value) - 1))
            hh_date = to_utc(hh_date_ct)
            price = Decimal(str(row[2].value))
            run = row[5].value

            try:
                rs, rates, rts = cache[hh_date.year][hh_date.month]
            except KeyError:
                _save_cache(sess, cache)

                try:
                    yr_cache = cache[hh_date.year]
                except KeyError:
                    yr_cache = cache[hh_date.year] = {}

                rs = sess.query(RateScript).filter(
                    RateScript.contract == contract,
                    RateScript.start_date <= hh_date, or_(
                        RateScript.finish_date == null(),
                        RateScript.finish_date >= hh_date)).first()
                while rs is None:
                    self.log(
                        "There's no rate script at " + hh_format(hh_date) +
                        ".")
                    latest_rs = sess.query(RateScript).filter(
                        RateScript.contract == contract).order_by(
                        RateScript.start_date.desc()).first()
                    contract.update_rate_script(
                        sess, latest_rs, latest_rs.start_date,
                        latest_rs.start_date + relativedelta(months=2) - HH,
                        loads(latest_rs.script))
                    new_rs_start = latest_rs.start_date + relativedelta(
                        months=1)
                    contract.insert_rate_script(sess, new_rs_start, {})
                    sess.commit()
                    self.log(
                        "Added a rate script starting at " +
                        hh_format(new_rs_start) + ".")

                    rs = sess.query(RateScript).filter(
                        RateScript.contract == contract,
                        RateScript.start_date <= hh_date, or_(
                            RateScript.finish_date == null(),
                            RateScript.finish_date >= hh_date)).first()

                rates = loads(rs.script)
                try:
                    rts = rates['rates_gbp_per_mwh']
                except KeyError:
                    rts = rates['rates_gbp_per_mwh'] = {}
                yr_cache[hh_date.month] = rs, rates, rts

            key = key_format(hh_date)
            try:
                existing = rts[key]
            except KeyError:
                existing = rts[key] = {}

            if run not in existing:
                existing[run] = price
                self.log(
                    "Added rate at " + hh_format(hh_date) + " for run " + run +
                    ".")

        _save_cache(sess, cache)
        book = sheet = None


def _discover_urls(logger):
    host = 'https://www.nationalgrideso.com'
    page = host + "/charging/balancing-services-use-system-bsuos-charges"
    logger("Searching for URLs on " + page)
    urls = set()
    res = requests.get(page)
    src = res.text
    for pref in (
            '" title="Current II BSUoS Data"',
            '" title="Current RF BSUoS Data"',
            '" title="Current SF BSUoS Data"'):
        pidx = src.find(pref)
        aidx = src.rfind('<', 0, pidx)
        urls.add(host + src[aidx+9:pidx])
    return urls


def get_importer():
    return bsuos_importer


def startup():
    global bsuos_importer
    bsuos_importer = BsuosImporter()
    bsuos_importer.start()


@atexit.register
def shutdown():
    if bsuos_importer is not None:
        bsuos_importer.stop()



from sqlalchemy import or_
from sqlalchemy.sql.expression import null
from sqlalchemy.orm import joinedload
import traceback
from chellow.models import Session, Contract, Era
import chellow.computer
from chellow.utils import (
    hh_min, hh_max, hh_format, req_int, csv_make_val, c_months_u)
from werkzeug.exceptions import BadRequest
import os
from flask import g
import threading
from chellow.views import chellow_redirect
import csv
import chellow.dloads


def content(contract_id, end_year, end_month, months, user):
    caches = {}
    sess = f = supply_source = None
    try:
        sess = Session()
        contract = Contract.get_dc_by_id(sess, contract_id)

        month_list = list(
            c_months_u(
                finish_year=end_year, finish_month=end_month, months=months))
        start_date, finish_date = month_list[0][0], month_list[-1][-1]

        forecast_date = chellow.computer.forecast_date()
        running_name, finished_name = chellow.dloads.make_names(
            'dc_virtual_bills.csv', user)

        f = open(running_name, mode='w', newline='')
        writer = csv.writer(f, lineterminator='\n')

        bill_titles = chellow.computer.contract_func(
            caches, contract, 'virtual_bill_titles')()
        header_titles = [
            'Import MPAN Core', 'Export MPAN Core', 'Start Date',
            'Finish Date']

        vb_func = chellow.computer.contract_func(
            caches, contract, 'virtual_bill')

        writer.writerow(header_titles + bill_titles)

        for era in sess.query(Era).distinct().filter(
                or_(
                    Era.finish_date == null(),
                    Era.finish_date >= start_date),
                Era.start_date <= finish_date,
                Era.dc_contract == contract).options(
                joinedload(Era.channels)).order_by(Era.supply_id):

            imp_mpan_core = era.imp_mpan_core
            if imp_mpan_core is None:
                imp_mpan_core_str = ''
                is_import = False
            else:
                is_import = True
                imp_mpan_core_str = imp_mpan_core

            exp_mpan_core = era.exp_mpan_core
            exp_mpan_core_str = '' if exp_mpan_core is None else exp_mpan_core

            chunk_start = hh_max(era.start_date, start_date)
            chunk_finish = hh_min(era.finish_date, finish_date)

            vals = [
                imp_mpan_core_str, exp_mpan_core_str, hh_format(chunk_start),
                hh_format(chunk_finish)]

            supply_source = chellow.computer.SupplySource(
                sess, chunk_start, chunk_finish, forecast_date, era, is_import,
                caches)
            vb_func(supply_source)
            bill = supply_source.dc_bill

            for title in bill_titles:
                vals.append(csv_make_val(bill.get(title)))
                if title in bill:
                    del bill[title]

            for k in sorted(bill.keys()):
                vals.append(k)
                vals.append(csv_make_val(bill[k]))

            writer.writerow(vals)

            # Avoid long-running transactions
            sess.rollback()
    except BadRequest as e:
        msg = 'Problem '
        if supply_source is not None:
            msg += "with supply " + supply_source.mpan_core + \
                " starting at " + hh_format(supply_source.start_date) + " "
        msg += str(e)
        writer.writerow([msg])
    except BaseException:
        msg = "Problem " + traceback.format_exc() + '\n'
        f.write(msg)
    finally:
        f.close()
        os.rename(running_name, finished_name)
        if sess is not None:
            sess.close()


def do_get(sess):
    end_year = req_int("end_year")
    end_month = req_int("end_month")
    months = req_int("months")
    contract_id = req_int('dc_contract_id')

    args = contract_id, end_year, end_month, months, g.user
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



import os
from datetime import datetime as Datetime

import chellow.bank_holidays
import chellow.bsuos
import chellow.dloads
import chellow.g_cv
import chellow.hh_importer
import chellow.rcrc
import chellow.system_price
import chellow.tlms
import chellow.utils
import chellow.views
from chellow.models import (
    Contract, Session, User, UserRole, db_upgrade, start_sqlalchemy)
from chellow.utils import to_ct, utc_datetime_now

from flask import Flask, Response, g, make_response, render_template, request

from werkzeug.exceptions import NotFound

from zish import dumps

from ._version import get_versions


TEMPLATE_FORMATS = {
    'year': '%Y', 'month': '%m', 'day': '%d', 'hour': '%H',
    'minute': '%M', 'full': '%Y-%m-%d %H:%M', 'date': '%Y-%m-%d'}


def create_app(testing=False):
    app = Flask('chellow', instance_relative_config=True)
    app.secret_key = os.urandom(24)
    start_sqlalchemy()

    app.register_blueprint(chellow.views.views)
    chellow.utils.root_path = app.root_path

    if not testing:
        db_upgrade(app.root_path)
        chellow.rcrc.startup()
        chellow.bsuos.startup()
        chellow.system_price.startup()
        chellow.hh_importer.startup()
        chellow.tlms.startup()
        chellow.bank_holidays.startup()
        chellow.dloads.startup(app.instance_path)
        chellow.g_cv.startup()

    @app.before_first_request
    def before_first_request():
        try:
            scheme = request.headers['X-Forwarded-Proto']
        except KeyError:
            sess = Session()
            try:
                config_contract = Contract.get_non_core_by_name(
                    sess, 'configuration')
                props = config_contract.make_properties()
                scheme = props.get('redirect_scheme', 'http')
            finally:
                sess.close()

        try:
            host = request.headers['X-Forwarded-Host']
        except KeyError:
            host = request.host

        chellow.utils.url_root = f'{scheme}://{host}/'

    @app.before_request
    def before_request():
        g.sess = Session()

        msg = ' '.join(
            '-' if v is None else v for v in (
                request.remote_addr, str(request.user_agent),
                request.remote_user,
                f"[{Datetime.now().strftime('%d/%b/%Y:%H:%M:%S')}]",
                f'"{request.method} {request.path} '
                f'{request.environ.get("SERVER_PROTOCOL")}"', None, None))
        print(msg)

    @app.before_request
    def check_permissions(*args, **kwargs):
        g.user = None
        config_contract = Contract.get_non_core_by_name(
            g.sess, 'configuration')
        props = config_contract.make_properties()
        ad_props = props.get('ad_authentication', {})
        ad_auth_on = ad_props.get('on', False)
        if ad_auth_on:
            username = request.headers['X-Isrw-Proxy-Logon-User'].upper()
            user = g.sess.query(User).filter(
                User.email_address == username).first()
            if user is None:
                try:
                    username = ad_props['default_user']
                    user = g.sess.query(User).filter(
                        User.email_address == username).first()
                except KeyError:
                    user = None
            if user is not None:
                g.user = user
        else:
            auth = request.authorization

            if auth is None:
                try:
                    ips = props['ips']
                    if request.remote_addr in ips:
                        key = request.remote_addr
                    elif '*.*.*.*' in ips:
                        key = '*.*.*.*'
                    else:
                        key = None

                    email = ips[key]
                    g.user = g.sess.query(User).filter(
                        User.email_address == email).first()
                except KeyError:
                    pass
            else:
                user = g.sess.query(User).filter(
                    User.email_address == auth.username).first()
                if user is not None and user.password_matches(auth.password):
                    g.user = user

        # Got our user
        path = request.path
        method = request.method
        if path in (
                '/health', '/nationalgrid/sf_bsuos.xls',
                '/nationalgrid/cv.csv',
                '/elexonportal/file/download/BESTVIEWPRICES_FILE', '/ecoes',
                '/elexonportal/file/download/TLM_FILE',
                '/elexonportal/file/download/RCRC_FILE',
                '/ecoes/NonDomesticCustomer/ExportPortfolioMPANs', '/hh_api'):
            return

        if g.user is not None:
            if 'X-Isrw-Proxy-Logon-User' in request.headers:
                g.user.proxy_username = \
                    request.headers['X-Isrw-Proxy-Logon-User'].upper()

            role = g.user.user_role
            role_code = role.code

            if role_code == "viewer" and (
                    method in ("GET", "HEAD") or path in (
                        '/reports/169', '/reports/187')):
                return
            elif role_code == "editor":
                return
            elif role_code == "party-viewer":
                if method in ("GET", "HEAD"):
                    party = g.user.party
                    market_role_code = party.market_role.code
                    if market_role_code in ('C', 'D'):
                        dc_contract_id = request.args["dc_contract_id"]
                        dc_contract = Contract.get_dc_by_id(
                            g.sess, dc_contract_id)
                        if dc_contract.party == party and \
                                request.full_path.startswith(
                                    "/channel_snags?"):
                            return
                    elif market_role_code == 'X':
                        if path.startswith("/supplier_contracts/" + party.id):
                            return

        if g.user is None and g.sess.query(User).count() == 0:
            g.sess.rollback()
            user_role = g.sess.query(UserRole).filter(
                UserRole.code == 'editor').one()
            User.insert(g.sess, 'admin@example.com', 'admin', user_role, None)
            g.sess.commit()
            return

        if g.user is None or (not ad_auth_on and auth is None):
            return Response(
                'Could not verify your access level for that URL.\n'
                'You have to login with proper credentials', 401,
                {'WWW-Authenticate': 'Basic realm="Chellow"'})

        config = Contract.get_non_core_by_name(g.sess, 'configuration')
        return make_response(
            render_template('403.html', properties=config.make_properties()),
            403)

    @app.context_processor
    def chellow_context_processor():
        global_alerts = []
        for task in chellow.hh_importer.tasks.values():
            if task.is_error:
                try:
                    contract = Contract.get_by_id(g.sess, task.contract_id)
                    global_alerts.append(
                        f"There's a problem with the automatic HH data "
                        f"importer for contract '{contract.name}'.")
                except NotFound:
                    pass

        for importer in (
                chellow.bsuos.bsuos_importer, chellow.g_cv.g_cv_importer):
            if importer is not None and importer.global_alert is not None:
                global_alerts.append(importer.global_alert)

        return {'current_user': g.user, 'global_alerts': global_alerts}

    @app.teardown_request
    def shutdown_session(exception=None):
        if getattr(g, 'sess', None) is not None:
            g.sess.close()

    @app.template_filter('hh_format')
    def hh_format_filter(dt, modifier='full'):
        if dt is None:
            return "Ongoing"
        else:
            return to_ct(dt).strftime(TEMPLATE_FORMATS[modifier])

    @app.template_filter('now_if_none')
    def now_if_none(dt):
        return utc_datetime_now() if dt is None else dt

    @app.template_filter('to_ct')
    def to_ct_filter(dt):
        return to_ct(dt)

    @app.template_filter('dumps')
    def dumps_filter(d):
        return dumps(d)

    return app


versions = get_versions()
__version__ = versions['version']
del get_versions



import traceback
import csv
from sqlalchemy.sql.expression import null
from werkzeug.exceptions import BadRequest
import requests
from itertools import chain
from chellow.models import Contract, Era, Supply, Source, Session, Party
from chellow.views import chellow_redirect
import chellow.dloads
import sys
import os
import threading
from flask import g


def content(user):
    sess = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names(
            'ecoes_comparison.csv', user)
        f = open(running_name, mode='w', newline='')
        writer = csv.writer(f, lineterminator='\n')

        props = Contract.get_non_core_by_name(sess, 'configuration'). \
            make_properties()

        ECOES_KEY = 'ecoes'
        try:
            ecoes_props = props[ECOES_KEY]
        except KeyError:
            raise BadRequest(
                "The property " + ECOES_KEY +
                " cannot be found in the configuration properties.")

        for key in ('user_name', 'password', 'prefix'):
            try:
                ecoes_props[key]
            except KeyError:
                raise BadRequest(
                    "The property " + key + " cannot be found in the " +
                    "'ecoes' section of the configuration properties.")

        proxies = props.get('proxies', {})
        s = requests.Session()
        r = s.get(ecoes_props['prefix'], proxies=proxies)
        r = s.post(
            ecoes_props['prefix'], data={
                'Username': ecoes_props['user_name'],
                'Password': ecoes_props['password'],
                }, allow_redirects=False)

        mpans = [
            v for (v,) in chain(
                sess.query(Era.imp_mpan_core).join(Supply, Source).
                join(Supply.dno).filter(
                    Party.dno_code.notin_(('88', '99')),
                    Era.finish_date == null(), Source.code != '3rd-party',
                    Era.imp_mpan_core != null()).distinct().order_by(
                    Era.imp_mpan_core),
                sess.query(Era.exp_mpan_core).join(Supply, Source)
                .join(Supply.dno).filter(
                    Party.dno_code.notin_(('88', '99')),
                    Era.finish_date == null(), Source.code != '3rd-party',
                    Era.exp_mpan_core != null()).distinct().order_by(
                    Era.exp_mpan_core))]
        r = s.get(
            ecoes_props['prefix'] +
            'NonDomesticCustomer/ExportPortfolioMPANs?fileType=csv',
            proxies=proxies)

        titles = (
            "MPAN Core", "MPAN Core No Spaces", "ECOES PC", "Chellow PC",
            "ECOES MTC", "Chellow MTC", "ECOES LLFC", "Chellow LLFC",
            "ECOES SSC", "Chellow SSC", "ECOES Supplier", "Chellow Supplier",
            "ECOES DC", "Chellow DC", "ECOES MOP", "Chellow MOP",
            "ECOES GSP Group", "Chellow GSP Group", "ECOES MSN", "Chellow MSN",
            "ECOES Meter Type", "Chellow Meter Type", "Problem"
        )
        writer.writerow(titles)

        parser = iter(csv.reader(r.text.splitlines(True)))
        next(parser)  # Skip titles

        for values in parser:
            problem = ''

            ecoes_titles = [
                'mpan-core', 'address-line-1', 'address-line-2',
                'address-line-3', 'address-line-4', 'address-line-5',
                'address-line-6', 'address-line-7', 'address-line-8',
                'address-line-9', 'post-code', 'supplier', 'registration-from',
                'mtc', 'mtc-date', 'llfc', 'llfc-from', 'pc', 'ssc',
                'measurement-class', 'energisation-status', 'da', 'dc', 'mop',
                'mop-appoint-date', 'gsp-group', 'gsp-effective-from', 'dno',
                'msn', 'meter-install-date', 'meter-type', 'map-id']

            ecoes = dict(zip(ecoes_titles, map(str.strip, values)))

            mpan_spaces = ' '.join(
                (
                    ecoes['mpan-core'][:2], ecoes['mpan-core'][2:6],
                    ecoes['mpan-core'][6:10], ecoes['mpan-core'][-3:]))

            ecoes_energisation_status = ecoes['energisation-status']
            disconnected = ecoes_energisation_status == ''
            current_chell = mpan_spaces in mpans

            if disconnected and current_chell:
                problem += "Disconnected in ECOES, but current in Chellow. "
            elif not disconnected and not current_chell:
                if ecoes_energisation_status == 'D':
                    st = "de-energised"
                else:
                    st = "energised"
                problem += "In ECOES (as " + st + ") but not current " + \
                    "in Chellow. "

            if current_chell:
                mpans.remove(mpan_spaces)
                supply = Supply.get_by_mpan_core(sess, mpan_spaces)
                era = supply.find_era_at(sess, None)
                if era.imp_mpan_core == mpan_spaces:
                    supplier_contract = era.imp_supplier_contract
                    llfc = era.imp_llfc
                else:
                    supplier_contract = era.exp_supplier_contract
                    llfc = era.exp_llfc

                chellow_pc = era.pc.code
                try:
                    if int(ecoes['pc']) != int(chellow_pc):
                        problem += "The PCs don't match. "
                except ValueError:
                    problem += "Can't parse the PC. "

                chellow_mtc = era.mtc.code
                try:
                    if int(ecoes['mtc']) != int(chellow_mtc):
                        problem += "The MTCs don't match. "
                except ValueError:
                    problem += "Can't parse the MTC. "

                chellow_llfc = llfc.code
                try:
                    if int(ecoes['llfc']) != int(chellow_llfc):
                        problem += "The LLFCs don't match. "
                except ValueError:
                    problem += "Can't parse the LLFC. "

                chellow_ssc = era.ssc
                if chellow_ssc is None:
                    chellow_ssc = ''
                    chellow_ssc_int = None
                else:
                    chellow_ssc = chellow_ssc.code
                    chellow_ssc_int = int(chellow_ssc)

                if len(ecoes['ssc']) > 0:
                    ecoes_ssc_int = int(ecoes['ssc'])
                else:
                    ecoes_ssc_int = None

                if ecoes_ssc_int != chellow_ssc_int and not (
                        ecoes_ssc_int is None and chellow_ssc_int is None):
                    problem += "The SSCs don't match. "

                chellow_supplier = supplier_contract.party.participant.code
                if chellow_supplier != ecoes['supplier']:
                    problem += "The supplier codes don't match. "

                dc_contract = era.dc_contract
                if dc_contract is None:
                    chellow_dc = ''
                else:
                    chellow_dc = dc_contract.party.participant.code

                if chellow_dc != ecoes['dc']:
                    problem += "The DC codes don't match. "

                mop_contract = era.mop_contract
                if mop_contract is None:
                    chellow_mop = ''
                else:
                    chellow_mop = mop_contract.party.participant.code

                if chellow_mop != ecoes['mop']:
                    problem += "The MOP codes don't match. "

                chellow_gsp_group = supply.gsp_group.code
                if chellow_gsp_group != ecoes['gsp-group']:
                    problem += "The GSP group codes don't match. "

                chellow_msn = era.msn
                if chellow_msn is None:
                    chellow_msn = ''
                if chellow_msn != ecoes['msn']:
                    problem += "The meter serial numbers don't match. "

                chellow_meter_type = _meter_type(era)

                if chellow_meter_type != ecoes['meter-type']:
                    problem += "The meter types don't match. See " \
                        "http://dtc.mrasco.com/DataItem.aspx?" \
                        "ItemCounter=0483 "
            else:
                chellow_pc = ''
                chellow_mtc = ''
                chellow_llfc = ''
                chellow_ssc = ''
                chellow_supplier = ''
                chellow_dc = ''
                chellow_mop = ''
                chellow_gsp_group = ''
                chellow_msn = ''
                chellow_meter_type = ''

            if len(problem) > 0:
                writer.writerow(
                    [
                        mpan_spaces, ecoes['mpan-core'], ecoes['pc'],
                        chellow_pc, ecoes['mtc'], chellow_mtc,
                        ecoes['llfc'], chellow_llfc, ecoes['ssc'],
                        chellow_ssc, ecoes['supplier'], chellow_supplier,
                        ecoes['dc'], chellow_dc, ecoes['mop'], chellow_mop,
                        ecoes['gsp-group'], chellow_gsp_group,
                        ecoes['msn'], chellow_msn, ecoes['meter-type'],
                        chellow_meter_type, problem])
            sess.expunge_all()

        for mpan_core in mpans:
            supply = Supply.get_by_mpan_core(sess, mpan_core)
            era = supply.find_era_at(sess, None)
            if era.imp_mpan_core == mpan_core:
                supplier_contract = era.imp_supplier_contract
                llfc = era.imp_llfc
            else:
                supplier_contract = era.exp_supplier_contract
                llfc = era.exp_llfc

            ssc = '' if era.ssc is None else era.ssc.code

            dc_contract = era.dc_contract
            if dc_contract is None:
                dc = ''
            else:
                dc = dc_contract.party.participant.code

            mop_contract = era.mop_contract
            if mop_contract is None:
                mop = ''
            else:
                mop = mop_contract.party.participant.code

            msn = '' if era.msn is None else era.msn

            meter_type = _meter_type(era)

            writer.writerow(
                [
                    mpan_core, mpan_core.replace(' ', ''), '', era.pc.code,
                    '', era.mtc.code, '', llfc.code, '', ssc, '',
                    supplier_contract.party.participant.code, '', dc, '',
                    mop, '', supply.gsp_group.code, '', msn, '',
                    meter_type, 'In Chellow, but not in ECOES.'])
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        writer.writerow([msg])
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def _meter_type(era):
    props = era.props
    try:
        return props['meter_type']
    except KeyError:
        if era.pc.code == '00':
            return 'H'
        elif len(era.channels) > 0:
            return 'RCAMR'
        elif era.mtc.meter_type.code in ['UM', 'PH']:
            return ''
        else:
            return 'N'


def do_get(sess):
    threading.Thread(target=content, args=(g.user,)).start()
    return chellow_redirect("/downloads", 303)



#!/usr/bin/env python

import sys
import time
from os import path

from setuptools import setup

import versioneer

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


if "--test" in sys.argv:
    versioneer.tstamp = str(int(time.time()))
    sys.argv.remove("--test")


setup(
    name='chellow',
    version=versioneer.get_version(),
    description='Web Application for checking UK utility bills.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Tony Locke',
    author_email='tlocke@tlocke.org.uk',
    url='https://github.com/WessexWater/chellow',
    cmdclass=versioneer.get_cmdclass(),
    packages=[
        'chellow', 'chellow.reports'
    ],
    package_data={
        'chellow': [
            'non_core_contracts/*/*.zish',
            'non_core_contracts/*/rate_scripts/*.zish',
            'mdd/converted/*.csv',
            'templates/*.html',
            'templates/css/*.css',
            'templates/js/*.js',
            'templates/*.css',
            'static/images/**',
            'static/css/**',
            'static/js/**',
            'static/font-awesome-4.6.3/**',
            'static/font-awesome-4.6.3/css/**',
            'static/font-awesome-4.6.3/fonts/**',
            'static/font-awesome-4.6.3/less/**',
            'static/font-awesome-4.6.3/scss/**',
            'nationalgrid/*',
            'elexonportal/*',
            'rate_scripts/*/*.zish'
        ]
    },
    install_requires=[
        'odio==0.0.18',
        'pg8000==1.13.2',
        'Flask==1.1.2',
        'SQLAlchemy==1.3.20',
        'openpyxl==2.4.8',
        'python-dateutil==2.8.1',
        'pytz==2020.1',
        'ftputil==4.0.0',
        'requests==2.24.0',
        'waitress==1.4.3',
        'pep3143daemon==0.0.6',
        'pip>=9.0.1',
        'pysftp==0.2.9',
        'pympler==0.9',
        'psutil==5.7.2',
        'xlrd==1.2.0',
        'zish==0.1.8'
    ],
    data_files=[('config', ['config/chellow.conf'])],
    entry_points={
        'console_scripts': [
            'chellow = chellow.commands:chellow_command',
            'chellow_test_setup = chellow.commands:chellow_test_setup'
        ]
    },
    scripts=['bin/chellow_service_monitor.sh', 'bin/chellow_start.sh']
)



import csv
import threading
import collections
import traceback
from dateutil.relativedelta import relativedelta
import requests
from chellow.models import (
    Session, Contract, RateScript, get_non_core_contract_id)
from chellow.utils import HH, hh_format, utc_datetime_now, to_utc, to_ct
from werkzeug.exceptions import BadRequest
import atexit
from datetime import datetime as Datetime
from decimal import Decimal, InvalidOperation
from sqlalchemy import or_
from sqlalchemy.sql.expression import null
from zish import loads, dumps


ELEXON_PORTAL_SCRIPTING_KEY_KEY = 'elexonportal_scripting_key'

RUNS = ['DF', 'RF', 'R3', 'R2', 'R1', 'SF', 'II']


def key_format(dt):
    return dt.strftime("%d %H:%M")


def hh(data_source, run='DF'):
    gsp_group_code = data_source.gsp_group_code
    use_runs = RUNS[RUNS.index(run):]

    try:
        cache = data_source.caches['tlms']
    except KeyError:
        cache = data_source.caches['tlms'] = {}

    for h in data_source.hh_data:
        try:
            h['tlm'] = tlm = cache[h['start-date']][gsp_group_code][run]
        except KeyError:
            h_start = h['start-date']
            db_id = get_non_core_contract_id('tlms')
            rates = data_source.hh_rate(db_id, h_start)['tlms']

            key = key_format(h_start)
            try:
                rate = rates[key]
            except KeyError:
                rate = sorted(rates.items())[-1][1]

            try:
                gsp_rate = rate[gsp_group_code]
            except KeyError:
                gsp_rate = sorted(rate.items())[-1][1]

            tlm = None
            for use_run in use_runs:
                try:
                    h['tlm'] = tlm = float(gsp_rate[use_run]['off_taking'])
                except KeyError:
                    pass

            if tlm is None:
                h['tlm'] = tlm = float(
                    sorted(gsp_rate.items())[-1][1]['off_taking'])

            try:
                rates_cache = cache[h['start-date']]
            except KeyError:
                rates_cache = cache[h['start-date']] = {}

            try:
                gsp_cache = rates_cache[gsp_group_code]
            except KeyError:
                gsp_cache = rates_cache[gsp_group_code] = {}

            gsp_cache[run] = tlm

        h['nbp-kwh'] = h['gsp-kwh'] * tlm


tlm_importer = None


class TlmImporter(threading.Thread):
    def __init__(self):
        super(TlmImporter, self).__init__(name="TLM Importer")
        self.lock = threading.RLock()
        self.messages = collections.deque(maxlen=100)
        self.stopped = threading.Event()
        self.going = threading.Event()
        self.PROXY_HOST_KEY = 'proxy.host'
        self.PROXY_PORT_KEY = 'proxy.port'

    def stop(self):
        self.stopped.set()
        self.going.set()
        self.join()

    def go(self):
        self.going.set()

    def is_locked(self):
        if self.lock.acquire(False):
            self.lock.release()
            return False
        else:
            return True

    def log(self, message):
        self.messages.appendleft(
            utc_datetime_now().strftime("%Y-%m-%d %H:%M:%S") + " - " + message)

    def run(self):
        while not self.stopped.isSet():
            if self.lock.acquire(False):
                try:
                    _import_tlms(self.log)
                finally:
                    self.lock.release()

            self.going.wait(24 * 60 * 60)
            self.going.clear()


def _save_cache(sess, cache):
    for yr, yr_cache in cache.items():
        for month, (rs, rates, rts) in tuple(yr_cache.items()):
            rs.script = dumps(rates)
            sess.commit()
            del yr_cache[month]


def _import_tlms(log_func):
    sess = None
    cache = {}
    try:
        sess = Session()
        log_func("Starting to check TLMs.")
        contract = Contract.get_non_core_by_name(sess, 'tlms')
        contract_props = contract.make_properties()
        if contract_props.get('enabled', False):

            config = Contract.get_non_core_by_name(sess, 'configuration')
            props = config.make_properties()
            scripting_key = props.get(ELEXON_PORTAL_SCRIPTING_KEY_KEY)
            if scripting_key is None:
                raise BadRequest(
                    "The property " + ELEXON_PORTAL_SCRIPTING_KEY_KEY +
                    " cannot be found in the configuration properties.")

            url_str = ''.join(
                (
                    contract_props['url'],
                    'file/download/TLM_FILE?key=', scripting_key))

            r = requests.get(url_str)
            parser = csv.reader(
                (x.decode() for x in r.iter_lines()),
                delimiter=',', quotechar='"')
            log_func("Opened " + url_str + ".")

            next(parser, None)
            for i, values in enumerate(parser):
                if values[3] == '':
                    for zone in GSP_GROUP_LOOKUP.keys():
                        values[3] = zone
                        _process_line(cache, sess, contract, log_func, values)
                else:
                    _process_line(cache, sess, contract, log_func, values)

            _save_cache(sess, cache)
        else:
            log_func(
                "The importer is disabled. Set 'enabled' to "
                "'true' in the properties to enable it.")

    except BadRequest as e:
        log_func("Problem: " + e.description)
        sess.rollback()
    except BaseException:
        log_func("Outer problem " + traceback.format_exc())
        sess.rollback()
    finally:
        if sess is not None:
            sess.close()
        log_func("Finished checking TLM rates.")


GSP_GROUP_LOOKUP = {
    '1': '_A',
    '2': '_B',
    '3': '_C',
    '4': '_D',
    '5': '_E',
    '6': '_F',
    '7': '_G',
    '8': '_H',
    '9': '_J',
    '10': '_K',
    '11': '_L',
    '12': '_M',
    '13': '_N',
    '14': '_P'
}


def _process_line(cache, sess, contract, log_func, values):
    hh_date_ct = to_ct(Datetime.strptime(values[0], "%d/%m/%Y"))
    hh_date = to_utc(hh_date_ct)
    hh_date += relativedelta(minutes=30*(int(values[2]) - 1))
    run = values[1]
    gsp_group_code = GSP_GROUP_LOOKUP[values[3]]
    off_taking_str = values[4]

    try:
        off_taking = Decimal(off_taking_str)
    except InvalidOperation as e:
        raise BadRequest(
            "Problem parsing 'off-taking' field '" + off_taking_str +
            "' in the row " + str(values) + ". " + str(e))

    delivering = Decimal(values[5])

    try:
        rs, rates, rts = cache[hh_date.year][hh_date.month]
    except KeyError:
        _save_cache(sess, cache)
        try:
            yr_cache = cache[hh_date.year]
        except KeyError:
            yr_cache = cache[hh_date.year] = {}

        rs = sess.query(RateScript).filter(
            RateScript.contract == contract,
            RateScript.start_date <= hh_date, or_(
                RateScript.finish_date == null(),
                RateScript.finish_date >= hh_date)).first()
        while rs is None:
            log_func("There's no rate script at " + hh_format(hh_date) + ".")
            latest_rs = sess.query(RateScript).filter(
                RateScript.contract == contract).order_by(
                RateScript.start_date.desc()).first()
            contract.update_rate_script(
                sess, latest_rs, latest_rs.start_date,
                latest_rs.start_date + relativedelta(months=2) - HH,
                loads(latest_rs.script))
            new_rs_start = latest_rs.start_date + relativedelta(months=1)
            contract.insert_rate_script(sess, new_rs_start, {})
            sess.commit()
            log_func(
                "Added a rate script starting at " + hh_format(new_rs_start) +
                ".")

            rs = sess.query(RateScript).filter(
                RateScript.contract == contract,
                RateScript.start_date <= hh_date, or_(
                    RateScript.finish_date == null(),
                    RateScript.finish_date >= hh_date)).first()

        rates = loads(rs.script)

        try:
            rts = rates['tlms']
        except KeyError:
            rts = rates['tlms'] = {}

        yr_cache[hh_date.month] = rs, rates, rts
        sess.rollback()

    key = key_format(hh_date)
    try:
        existing = rts[key]
    except KeyError:
        existing = rts[key] = {}

    try:
        group = existing[gsp_group_code]
    except KeyError:
        group = existing[gsp_group_code] = {}

    if run not in group:
        group[run] = {
            'off_taking': off_taking,
            'delivering': delivering}

        log_func(
            "Found rate at " + hh_format(hh_date) + " for GSP Group " +
            gsp_group_code + " and run " + run + ".")


def get_importer():
    return tlm_importer


def startup():
    global tlm_importer
    if tlm_importer is not None:
        raise BadRequest("The TLM importer has already been started.")
    tlm_importer = TlmImporter()
    tlm_importer.start()


@atexit.register
def shutdown():
    if tlm_importer is not None:
        tlm_importer.stop()



from dateutil.relativedelta import relativedelta
import traceback
from sqlalchemy import or_
from sqlalchemy.sql.expression import null, true
from chellow.computer import contract_func, forecast_date
from chellow.g_engine import GDataSource
from chellow.models import Session, GContract, GEra, Site, SiteGEra
import chellow.dloads
import csv
from chellow.utils import (
    HH, hh_format, req_date, req_int, make_val, hh_max, hh_min, utc_datetime)
import sys
import os
import threading
from flask import g
from chellow.views import chellow_redirect
from werkzeug.exceptions import BadRequest


def content(start_date, finish_date, g_contract_id, user):
    report_context = {}
    sess = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names(
            'gas_virtual_bills.csv', user)
        f = open(running_name, mode='w', newline='')
        writer = csv.writer(f, lineterminator='\n')

        g_contract = GContract.get_by_id(sess, g_contract_id)
        forecast_dt = forecast_date()

        month_start = utc_datetime(start_date.year, start_date.month, 1)
        month_finish = month_start + relativedelta(months=1) - HH

        bill_titles = contract_func(
            report_context, g_contract, 'virtual_bill_titles')()
        writer.writerow(
            [
                'MPRN', 'Site Code', 'Site Name', 'Account', 'From', 'To'] +
            bill_titles)

        while not month_start > finish_date:
            period_start = hh_max(start_date, month_start)
            period_finish = hh_min(finish_date, month_finish)

            for g_era in sess.query(GEra).distinct().filter(
                    GEra.g_contract == g_contract,
                    GEra.start_date <= period_finish, or_(
                        GEra.finish_date == null(),
                        GEra.finish_date >= period_start)):

                chunk_start = hh_max(g_era.start_date, period_start)
                chunk_finish = hh_min(g_era.finish_date, period_finish)

                data_source = GDataSource(
                    sess, chunk_start, chunk_finish, forecast_dt, g_era,
                    report_context, None)

                site = sess.query(Site).join(SiteGEra).filter(
                    SiteGEra.g_era == g_era,
                    SiteGEra.is_physical == true()).one()

                vals = [
                    data_source.mprn, site.code, site.name,
                    data_source.account,
                    hh_format(data_source.start_date),
                    hh_format(data_source.finish_date)]

                contract_func(report_context, g_contract, 'virtual_bill')(
                    data_source)
                bill = data_source.bill
                for title in bill_titles:
                    if title in bill:
                        val = make_val(bill[title])
                        del bill[title]
                    else:
                        val = ''
                    vals.append(val)

                for k in sorted(bill.keys()):
                    vals.append(k)
                    vals.append(str(bill[k]))
                writer.writerow(vals)

            month_start += relativedelta(months=1)
            month_finish = month_start + relativedelta(months=1) - HH
    except BadRequest as e:
        writer.writerow(["Problem: " + e.description])
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        writer.writerow([msg])
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    start_date = req_date('start')
    finish_date = req_date('finish')
    g_contract_id = req_int('g_contract_id')

    threading.Thread(
        target=content, args=(start_date, finish_date, g_contract_id, g.user)
        ).start()
    return chellow_redirect("/downloads", 303)



from decimal import Decimal
import decimal
from datetime import datetime as Datetime
from chellow.utils import parse_mpan_core, to_utc, to_ct
from xlrd import xldate_as_tuple, open_workbook
from werkzeug.exceptions import BadRequest
from chellow.models import Session


def get_date(row, name, datemode):
    val = get_value(row, name)
    if isinstance(val, float):
        return to_utc(to_ct(Datetime(*xldate_as_tuple(val, datemode))))

    raise BadRequest("Can't find a date.")


def get_value(row, idx):
    try:
        return row[idx].value
    except IndexError:
        raise BadRequest(
            "For the row " + str(row) + ", the index is " + str(idx) +
            " which is beyond the end of the row. ")


def get_str(row, idx):
    return get_value(row, idx).strip()


def get_dec(row, idx):
    try:
        return Decimal(str(get_value(row, idx)))
    except decimal.InvalidOperation:
        return None


def get_int(row, idx):
    return int(get_value(row, idx))


class Parser():
    def __init__(self, f):
        self.book = open_workbook(file_contents=f.read())
        self.sheet = self.book.sheet_by_index(1)

        self.last_line = None
        self._line_number = None
        self._title_line = None

    @property
    def line_number(self):
        return None if self._line_number is None else self._line_number + 1

    def _set_last_line(self, i, line):
        self._line_numer = i
        self.last_line = line
        if i == 0:
            self._title_line = line
        return line

    def make_raw_bills(self):
        row_index = sess = None
        try:
            sess = Session()
            bills = []
            title_row = self.sheet.row(10)
            issue_date = get_date(self.sheet.row(5), 2, self.book.datemode)
            if issue_date is None:
                raise BadRequest("Expected to find the issue date at cell C6.")

            for row_index in range(11, self.sheet.nrows):
                row = self.sheet.row(row_index)
                val = get_value(row, 1)
                if val is None or val == '':
                    break

                self._set_last_line(row_index, val)
                mpan_core = parse_mpan_core(str(get_int(row, 1)))

                start_date = finish_date = get_date(row, 5, self.book.datemode)
                activity_name_raw = get_str(row, 6)
                activity_name = activity_name_raw.lower().replace(' ', '_')

                net_dec = get_dec(row, 8)
                if net_dec is None:
                    raise BadRequest(
                        "Can't find a decimal at column I, expecting the net "
                        "GBP.")
                net = round(net_dec, 2)

                vat_dec = get_dec(row, 9)
                if vat_dec is None:
                    raise BadRequest(
                        "Can't find a decimal at column J, expecting the VAT "
                        "GBP.")

                vat = round(vat_dec, 2)

                gross_dec = get_dec(row, 10)
                if gross_dec is None:
                    raise BadRequest(
                        "Can't find a decimal at column K, expecting the "
                        "gross GBP.")

                gross = round(gross_dec, 2)

                breakdown = {
                    'raw-lines': [str(title_row)],
                    'activity-name': [activity_name], 'activity-gbp': net
                }

                bills.append(
                    {
                        'bill_type_code': 'N', 'kwh': Decimal(0), 'vat': vat,
                        'net': net, 'gross': gross, 'reads': [],
                        'breakdown': breakdown, 'account': mpan_core,
                        'issue_date': issue_date, 'start_date': start_date,
                        'finish_date': finish_date, 'mpan_core': mpan_core,
                        'reference': '_'.join(
                            (
                                start_date.strftime('%Y%m%d'),
                                finish_date.strftime('%Y%m%d'),
                                issue_date.strftime('%Y%m%d'),
                                mpan_core
                            )
                        )
                    }
                )
        except BadRequest as e:
            raise BadRequest(
                "Row number: " + str(row_index) + " " + e.description)
        finally:
            if sess is not None:
                sess.close()

        return bills
